"""
Beta from Phase
--------------------

This module contains some of the beta calculation related functionality of ``optics_measurements``.
It provides functions to calculate beta and alpha functions from phase advance data.
"""
import os
import re

import numpy as np
import pandas as pd
import tfs
from scipy.linalg import circulant

from omc3 import __version__ as VERSION
from omc3.definitions.constants import PI2
from omc3.optics_measurements.constants import (BETA_NAME, DELTA, ERR, EXT,
                                                MDL)
from omc3.optics_measurements.toolbox import df_diff, df_ratio, df_rel_diff
from omc3.utils import logging_tools, stats

LOGGER = logging_tools.get_logger(__name__)

EPSILON = 1.0E-16
ZERO_THRESHOLD = 1e-2
COT_THRESHOLD = 15.9
RCOND = 1.0e-14

METH_3BPM = "3BPM method"
METH_A_NBPM = "Analytical N-BPM method"
METH_NO_ERR = "No Errors"


def calculate(meas_input, tunes, phase_dict, header_dict, plane):
    """
    Calculates betas and alphas from phase advances.

    Args:
        meas_input: `OpticsInput` object
        tunes: `TuneDict` contains measured tunes.
        phase_dict: contains measured phase advances.
        header_dict:  dictionary of header items common for all output files.
        plane: marking the horizontal or vertical plane, **X** or **Y**.

    Returns:
        `BetaDict` object containing specific `TfsDataFrame` with results.
    """
    meas_and_model_tunes = (tunes[plane]["Q"], tunes[plane]["QM"] % 1) \
        if meas_input.compensation == "none" else (tunes[plane]["QF"], tunes[plane]["QFM"] % 1)

    if meas_input.three_bpm_method:
        beta_df = three_bpm_method(meas_input, phase_dict['free'], plane, meas_and_model_tunes)
        error_method = METH_3BPM
    else:
        beta_df, error_method = n_bpm_method(meas_input, phase_dict['free'], plane, meas_and_model_tunes)
    LOGGER.info(f"Errors from {error_method}")
    rmsbb = stats.weighted_rms(beta_df.loc[:, f"{DELTA}BET{plane}"].to_numpy(), errors=beta_df.loc[:, f"{ERR}{DELTA}BET{plane}"].to_numpy()) * 100
    LOGGER.info(f" - RMS beta beat: {rmsbb:.3f}%")
    header = _get_header(header_dict, error_method, meas_input.range_of_bpms, rmsbb)
    return beta_df, header


def write(beta_df, header, outputdir, plane):
    tfs.write(os.path.join(outputdir, f"{BETA_NAME}{plane.lower()}{EXT}"), beta_df,
              header, save_index="NAME")


def n_bpm_method(meas_input, phase, plane, meas_and_mdl_tunes):
    """
    Calculates betas and alphas from using all BPM combination within **range_of_bpms**. It also
    accounts for systematic errors

    Args:
        meas_input: Optics measurement configuration object.
        phase: phase matrices of measurement with errors and model tfs (bpm x bpm).
        plane: marking the horizontal or vertical plane, **X** or **Y**.
        meas_and_mdl_tunes: measured and model tunes.

    Returns:
        `TfsDataFrame` containing betas and alfas from phase.
    """
    elements, error_method = get_elements_with_errors(meas_input, plane)
    beta_df = _get_filtered_model_df(meas_input, phase, plane)
    bk_model = _get_filtered_model_df(meas_input, phase, plane, best=True)
    tune, mdltune = meas_and_mdl_tunes
    betas_alfas = np.zeros((len(phase["MEAS"].index), 4))
    nbpms = len(bk_model.index)
    n_comb = np.zeros(nbpms, dtype=int)
    m = int(meas_input.range_of_bpms / 2)
    loc_range = np.arange(-m, m + 1)
    phases_meas = phase["MEAS"] * PI2
    phases_err = phase["ERRMEAS"] * PI2
    phases_err.where(phases_err.notnull(), 1, inplace=True)

    for indx, probed_bpm_name in enumerate(bk_model.index):
        indx_el_first = elements.index.get_loc(bk_model.index[(indx - m) % nbpms])
        indx_el_last = elements.index.get_loc(bk_model.index[(indx + m) % nbpms])
        mu_column = "MU" + plane
        if indx < m:
            outer_meas_phase_adv = pd.concat((phases_meas.iloc[indx, nbpms + indx - m:] - tune * PI2, phases_meas.iloc[indx, :indx + m + 1]))
            outer_meas_err = pd.concat((phases_err.iloc[indx, nbpms + indx - m:], phases_err.iloc[indx, :indx + m + 1]))
            outer_mdl_ph = np.concatenate((bk_model.iloc[nbpms + indx - m:][mu_column] - mdltune, bk_model.iloc[:indx + m + 1][mu_column])) * PI2
            outer_elmts = pd.concat((elements.iloc[indx_el_first:], elements.iloc[:indx_el_last + 1]))
            outer_elmts_ph = np.concatenate((elements.iloc[indx_el_first:][mu_column] - mdltune, elements.iloc[:indx_el_last + 1][mu_column])) * PI2
        elif indx + m >= nbpms:
            outer_meas_phase_adv = pd.concat((phases_meas.iloc[indx, indx - m:], phases_meas.iloc[indx, :indx + m + 1 - nbpms] + tune * PI2))
            outer_meas_err = pd.concat((phases_err.iloc[indx, indx - m:], phases_err.iloc[indx, :indx + m + 1 - nbpms]))
            outer_mdl_ph = np.concatenate((bk_model.iloc[indx - m:][mu_column], bk_model.iloc[:indx + m + 1 - nbpms][mu_column] + mdltune)) * PI2
            outer_elmts = pd.concat((elements.iloc[indx_el_first:], elements.iloc[:indx_el_last + 1]))
            outer_elmts_ph = np.concatenate((elements.iloc[indx_el_first:][mu_column], elements.iloc[:indx_el_last + 1][mu_column] + mdltune)) * PI2
        else:
            outer_meas_phase_adv = phases_meas.iloc[indx, indx + loc_range]
            outer_meas_err = phases_err.iloc[indx, indx + loc_range]
            outer_mdl_ph = bk_model.iloc[indx + loc_range][mu_column].to_numpy() * PI2
            outer_elmts = elements.iloc[indx_el_first:indx_el_last + 1]
            outer_elmts_ph = elements.iloc[indx_el_first:indx_el_last + 1][mu_column].to_numpy() * PI2
        # bpms_inds_elements was not used (jdilly)
        # bpms_inds_elements = [outer_elmts.index.get_loc(bpm_name) for bpm_name in outer_meas_phase_adv.index.to_numpy()]
        sin_squared_elements = np.square(np.sin(
            outer_elmts_ph[:, np.newaxis] - outer_mdl_ph[np.newaxis, :]
        ))
        with np.errstate(divide='ignore'):
            cot_meas = 1.0 / np.tan(outer_meas_phase_adv.to_numpy())
            cot_model = 1.0 / np.tan((outer_mdl_ph - outer_mdl_ph[m]))
        patter = (np.abs(cot_meas) <= COT_THRESHOLD) & (np.abs(cot_model) <= COT_THRESHOLD)
        diag = np.concatenate((np.square(outer_meas_err.to_numpy()), outer_elmts.loc[:]["dK1"],
                               outer_elmts.loc[:]["dX"], outer_elmts.loc[:]["KdS"],
                               outer_elmts.loc[:]["mKdS"]))
        outer_elmts = outer_elmts.rename(columns={"BET" + plane: "BETA"})
        index_tuples = [[x, y] for x in loc_range[patter] + m for y in loc_range[patter] + m
                        if (x < y) and (abs(cot_model[x] - cot_model[y]) > ZERO_THRESHOLD) and
                        (np.sign(cot_model[x] - cot_model[y]) * np.sign(cot_meas[x] - cot_meas[y]) > 0)]
        mat_t_beta, mat_t_alpha = np.zeros((len(index_tuples), len(diag))), np.zeros((len(index_tuples), len(diag)))
        betas, alphas = np.empty(len(index_tuples)), np.empty(len(index_tuples))
        for i, c in enumerate(index_tuples):
            betas[i], alphas[i], mat_t_beta[i], mat_t_alpha[i] = \
                calculate_beta_alpha_from_single_combination(c, sin_squared_elements, outer_elmts, cot_model,
                                                             cot_meas, outer_meas_phase_adv, probed_bpm_name,
                                                             bk_model.at[probed_bpm_name, "BET" + plane],
                                                             bk_model.at[probed_bpm_name, "ALF" + plane], meas_input.range_of_bpms)

        mask = diag != 0
        mat_diag = np.diag(diag[mask])
        mat_v_beta = np.dot(mat_t_beta[:, mask], np.dot(mat_diag, np.transpose(mat_t_beta[:, mask])))
        mat_v_alpha = np.dot(mat_t_alpha[:, mask], np.dot(mat_diag, np.transpose(mat_t_alpha[:, mask])))

        if np.any(mat_v_beta) and np.any(mat_v_alpha):
            beti, beterr = _covariant_weighting(mat_v_beta, betas)
            alfi, alferr = _covariant_weighting(mat_v_alpha, alphas)
            n_comb[indx] = len(betas)
        else:
            LOGGER.debug(f"ValueError or no combinations left at {probed_bpm_name}.")
            LOGGER.debug(f"betas:\n{betas}")
            LOGGER.debug(f"alphas:\n{alphas}")
            continue
        betas_alfas[indx, :] = np.array([beti, beterr, alfi, alferr])

    beta_df[f"BET{plane}"] = betas_alfas[:, 0]
    beta_df[f"{ERR}BET{plane}"] = betas_alfas[:, 1]
    beta_df[f"ALF{plane}"] = betas_alfas[:, 2]
    beta_df[f"{ERR}ALF{plane}"] = betas_alfas[:, 3]
    beta_df["NCOMB"] = n_comb
    LOGGER.debug(f"No valid combinations for BPMs: {list(beta_df.index[beta_df['NCOMB'] == 0])}.")
    beta_df = beta_df.loc[beta_df["NCOMB"] > 0]
    beta_df = beta_df.loc[beta_df[f"BET{plane}"] > 0]
    too_high_error_mask = np.logical_or(beta_df[f"BET{plane}"] > beta_df[f"{ERR}BET{plane}"],
                                        beta_df[f"BET{plane}{MDL}"] > beta_df[f"{ERR}BET{plane}"])
    beta_df = beta_df.loc[too_high_error_mask]
    beta_df = _get_delta_columns(beta_df, plane)
    return beta_df, error_method


def get_elements_with_errors(meas_input, plane):
    if meas_input.accelerator.error_defs_file is None:
        raise IOError(f"Error definition file could not be found")
    elements = meas_input.accelerator.elements.loc[:, ["S", "K1L", "K2L", f"MU{plane}", f"BET{plane}"]]
    LOGGER.debug("Accelerator Error Definition")
    elements = _assign_uncertainties(elements, meas_input.accelerator.error_defs_file)
    errors_assigned = (len(elements["dK1"].to_numpy().nonzero()[0]) + len(
        elements["dX"].to_numpy().nonzero()[0]) + len(elements["KdS"].to_numpy().nonzero()[0])) > 0
    if not errors_assigned:
        LOGGER.warning("No systematic errors were given or no element was found for the given "
                       "error definitions. The systematic lattice errors are not used.")
    error_method = METH_A_NBPM if errors_assigned else METH_NO_ERR
    return elements, error_method


def calculate_beta_alpha_from_single_combination(c, sin_squared_elements, outer_elmts, cot_model,
                                                 cot_meas, outer_meas_phase_adv, probed_bpm_name,
                                                 betmdl1, alfmdl1, range_of_bpms):
    """
    Calculates beta and alpha functions as well as the respective covariance matrix lines for the
    given BPM combination (triplet).

    Args:
        c: relative indices of other two BPMs wrt probed one.
        sin_squared_elements:
        outer_elmts:
        cot_model:
        cot_meas:
        outer_meas_phase_adv:
        probed_bpm_name:
        betmdl1:
        alfmdl1:
        range_of_bpms:

    Returns:

    """
    m = int(range_of_bpms / 2)
    ix = c[0]
    iy = c[1]
    fac1, fac2 = -np.sign(c[0]-m), np.sign(c[1]-m)
    dif_cot_model = cot_model[ix] - cot_model[iy]
    # calculate beta
    dif_cot_meas = cot_meas[ix] - cot_meas[iy]
    denom = dif_cot_model / betmdl1
    beta_i = dif_cot_meas / denom
    avg_cot_model = (cot_model[ix] + cot_model[iy]) / 2
    denomalf = 2 * (avg_cot_model + alfmdl1)
    avg_cot_meas = (cot_meas[ix] + cot_meas[iy]) / 2

    alfa_i = 0.5 * (denomalf * dif_cot_meas / dif_cot_model - 2 * avg_cot_meas)

    lng = len(outer_elmts)
    line_length = 4 * len(outer_elmts.index) + 2 * m + 1
    outer_elmts_bet = outer_elmts.loc[:, "BETA"].to_numpy()
    outer_el_k2 = outer_elmts.loc[:, "K2L"].to_numpy()
    betaline = np.zeros(line_length)
    alfaline = np.zeros(line_length)

    # slice
    mloc = outer_elmts.index.get_loc(probed_bpm_name)
    xloc_r = outer_elmts.index.get_loc(outer_meas_phase_adv.index[ix])
    yloc_r = outer_elmts.index.get_loc(outer_meas_phase_adv.index[iy])
    denom_sinx = sin_squared_elements[xloc_r, m]
    denom_siny = sin_squared_elements[yloc_r, m]
    xmloc1, xmloc2 = min(xloc_r, mloc), max(xloc_r, mloc)
    ymloc1, ymloc2 = min(yloc_r, mloc), max(yloc_r, mloc)

    # get betas and sin for the elements in the slice
    elem_ph_xa = sin_squared_elements[xmloc1:xmloc2, ix]
    elem_ph_ya = sin_squared_elements[ymloc1:ymloc2, iy]
    elem_beta_xa = outer_elmts_bet[xmloc1:xmloc2]
    elem_beta_ya = outer_elmts_bet[ymloc1:ymloc2]
    elem_k2_xa = outer_el_k2[xmloc1:xmloc2]
    elem_k2_ya = outer_el_k2[ymloc1:ymloc2]

    bet_sin_ix = elem_ph_xa * elem_beta_xa / (denom_sinx * denom)
    bet_sin_iy = elem_ph_ya * elem_beta_ya / (denom_siny * denom)

    off1 = range_of_bpms
    off2 = range_of_bpms + lng
    off3 = range_of_bpms + 2 * lng
    off4 = range_of_bpms + 3 * lng

    # apply phase uncertainty
    betaline[ix] = -1 / (denom_sinx * denom)
    betaline[iy] = 1 / (denom_siny * denom)
    # apply quadrupolar field uncertainty (quadrupole longitudinal misalignment already included)
    betaline[xmloc1 + off1:xmloc2 + off1] += fac1 * bet_sin_ix
    betaline[ymloc1 + off1:ymloc2 + off1] += fac2 * bet_sin_iy
    # apply sextupole transverse misalignment
    betaline[xmloc1 + off2: xmloc2 + off2] += fac1 * elem_k2_xa * bet_sin_ix
    betaline[ymloc1 + off2: ymloc2 + off2] += fac2 * elem_k2_ya * bet_sin_iy
    # apply quadrupole longitudinal misalignments
    betaline[xmloc1 + off3: xmloc2 + off3] += fac1 * bet_sin_ix
    betaline[ymloc1 + off3: ymloc2 + off3] += fac2 * bet_sin_iy
    betaline[xmloc1 + off4: xmloc2 + off4] -= fac1 * bet_sin_ix
    betaline[ymloc1 + off4: ymloc2 + off4] -= fac2 * bet_sin_iy
    # apply phase uncertainty
    alfaline[ix] = -1 / (denom_sinx * denom * betmdl1) * denomalf + 1 / denom_sinx
    alfaline[iy] = 1 / (denom_siny * denom * betmdl1) * denomalf + 1 / denom_siny
    # apply quadrupolar field uncertainty (quadrupole longitudinal misalignment already included)
    alfaline[xmloc1 + off1:xmloc2 + off1] += fac1 * (.5 * (bet_sin_ix * denomalf + bet_sin_ix / betmdl1 * dif_cot_meas))
    alfaline[ymloc1 + off1:ymloc2 + off1] += fac2 * (.5 * (bet_sin_iy * denomalf + bet_sin_iy / betmdl1 * dif_cot_meas))
    # apply sextupole transverse misalignment
    alfaline[xmloc1 + off2: xmloc2 + off2] += fac1 * elem_k2_xa * bet_sin_ix
    alfaline[ymloc1 + off2: ymloc2 + off2] += fac2 * elem_k2_ya * bet_sin_iy
    # apply quadrupole longitudinal misalignments
    alfaline[xmloc1 + off3: xmloc2 + off3] += fac1 * (.5 * elem_k2_xa * (bet_sin_ix * denomalf + bet_sin_ix / betmdl1 * dif_cot_meas))
    alfaline[ymloc1 + off3: ymloc2 + off3] += fac2 * (.5 * elem_k2_ya * (bet_sin_iy * denomalf + bet_sin_iy / betmdl1 * dif_cot_meas))
    alfaline[xmloc1 + off4: xmloc2 + off4] -= fac1 * (.5 * (bet_sin_ix * denomalf + bet_sin_ix / betmdl1 * dif_cot_meas))
    alfaline[ymloc1 + off4: ymloc2 + off4] -= fac2 * (.5 * (bet_sin_iy * denomalf + bet_sin_iy / betmdl1 * dif_cot_meas))

    return beta_i, alfa_i, betaline, alfaline


def _covariant_weighting(mat, col):
    mat_inv = np.linalg.pinv(mat, rcond=RCOND)
    wb = np.sum(mat_inv, axis=1)
    mat_inv_sum = np.sum(wb)
    if mat_inv_sum == 0:
        raise ValueError
    # returns value and error
    return float(np.dot(wb.T, col) / mat_inv_sum), np.sqrt(np.dot(wb.T, np.dot(mat, wb)) / mat_inv_sum ** 2)


def _assign_uncertainties(twiss_full, errordefspath):
    """
    Adds uncertainty information to ``twiss_full``.

    Sources of Errors:
        ``dK1``:    quadrupolar field errors.
        ``dS``:     quadrupole longitudinal misalignments.
        ``dX``:     sextupole transverse misalignments.
        ``BPMdS``:  BPM longitudinal misalignments.
    """
    LOGGER.debug("Start creating uncertainty information")
    errdefs = tfs.read(errordefspath)
    twiss_full = twiss_full.assign(UNC=False, dK1=0, KdS=0, mKdS=0, dX=0, BPMdS=0)
    # loop over uncertainty definitions, fill the respective columns, set UNC to true
    for indx in errdefs.index:
        patt = errdefs.loc[indx, "PATTERN"]
        if patt.startswith("key:"):
            LOGGER.debug(f"creating uncertainty information for {patt}")
            mask = patt.split(":")[1]
        else:
            reg = re.compile(patt)
            LOGGER.debug(f"creating uncertainty information for RegEx {patt}")
            mask = twiss_full.index.str.contains(reg)

        twiss_full.loc[mask, "dK1"] = (errdefs.loc[indx, "dK1"] * twiss_full.loc[mask, "K1L"]) ** 2
        twiss_full.loc[mask, "dX"] = errdefs.loc[indx, "dX"]*2
        if errdefs.loc[indx, "MAINFIELD"] == "BPM":
            twiss_full.loc[mask, "BPMdS"] = errdefs.loc[indx, "dS"]**2
        else:
            twiss_full.loc[mask, "KdS"] = (errdefs.loc[indx, "dS"] * twiss_full.loc[mask, "K1L"]) ** 2
        twiss_full.loc[mask, "UNC"] = True

    # in case of quadrupole longitudinal misalignments, the element (DRIFT) in front of the
    # misaligned quadrupole will be used for the thin lens approximation of the misalignment
    twiss_full["mKdS"] = np.roll(twiss_full.loc[:]["KdS"], 1)
    twiss_full.loc[:, "UNC"] = np.logical_or(abs(np.roll(twiss_full.loc[:, "dK1"], -1)) > 1.0e-12,
                                             twiss_full.loc[:, "UNC"])
    LOGGER.debug("DONE creating uncertainty information")
    return twiss_full.loc[twiss_full["UNC"]]


def _get_header(header_dict, error_method, range_of_bpms, rmsbb):
    header = header_dict.copy()
    header['BetaAlgorithmVersion'] = VERSION
    header['RCond'] = RCOND
    header['RangeOfBPMs'] = "Adjacent" if error_method == METH_3BPM else range_of_bpms
    header['ErrorsFrom:'] = error_method
    header["RMS_BETABEAT"] = f"{rmsbb:.3f} %"
    return header


def three_bpm_method(meas_input, phase, plane, meas_and_mdl_tunes):
    """
    Calculates betas and alphas from using adjacent BPMs (3 combination).
    ``phase["MEAS"]``, ``phase["MODEL"]``, ``phase["ERRMEAS"]`` (from ``get_phases``) are of the
    form:

    +----------+----------+----------+----------+----------+
    |          |   BPM1   |   BPM2   |   BPM3   |   BPM4   |
    +----------+----------+----------+----------+----------+
    |   BPM1   |    0     |  phi_21  |  phi_31  |  phi_41  |
    +----------+----------+----------+----------+----------+
    |   BPM2   |  phi_12  |     0    |  phi_32  |  phi_42  |
    +----------+----------+----------+----------+----------+
    |   BPM3   |  phi_13  |  phi_23  |    0     |  phi_43  |
    +----------+----------+----------+----------+----------+

    and ``tilt_slice_matrix(matrix, shift, slice, tune)`` brings it into the form:

    +-----------+--------+--------+--------+--------+
    |           |  BPM1  |  BPM2  |  BPM3  |  BPM4  |
    +-----------+--------+--------+--------+--------+
    | BPM_(i-1) | phi_1n | phi_21 | phi_32 | phi_43 |
    +-----------+--------+--------+--------+--------+
    | BPM_i     |    0   |    0   |    0   |    0   |
    +-----------+--------+--------+--------+--------+
    | BPM_(i+1) | phi_12 | phi_23 | phi_34 | phi_45 |
    +-----------+--------+--------+--------+--------+

    ``cot_phase_*_shift1``:

    +-----------------------------+-----------------------------+-----------------------------+
    | cot(phi_1n) - cot(phi_1n-1) |  cot(phi_21) - cot(phi_2n)  |   cot(phi_32) - cot(phi_31) |
    +-----------------------------+-----------------------------+-----------------------------+
    |         NaN                 |         NaN                 |         NaN                 |
    +-----------------------------+-----------------------------+-----------------------------+
    |         NaN                 |         NaN                 |         NaN                 |
    +-----------------------------+-----------------------------+-----------------------------+
    |  cot(phi_13) - cot(phi_12)  |  cot(phi_24) - cot(phi_23)  |   cot(phi_35) - cot(phi_34) |
    +-----------------------------+-----------------------------+-----------------------------+

    - for the combination xxxABBx: first row,
    - for the combinstion xBBAxxx: fourth row,
    - for the combination xxBABxx: second row of ``cot_phase_*_shift2``.

    Args:
        meas_input: Optics measurement configuration object.
        phase: phase matrices of measurement with errors and model tfs (bpm x bpm).
        plane: marking the horizontal or vertical plane, **X** or **Y**.
        meas_and_mdl_tunes: measured  and model tunes.

    Returns:
        `TfsDataFrame` containing betas and alfas from phase.
    """
    tune, mdltune = meas_and_mdl_tunes
    beta_df = _get_filtered_model_df(meas_input, phase, plane)
    # tilt phase advances in order to have the phase advances in a neighbourhood
    tilted_meas = _tilt_slice_matrix(phase["MEAS"].to_numpy(), 2, 5, tune) * PI2
    tilted_model = _tilt_slice_matrix(phase["MODEL"].to_numpy(), 2, 5, mdltune) * PI2
    tilted_errmeas = _tilt_slice_matrix(phase["ERRMEAS"].to_numpy(), 2, 5, mdltune) * PI2
    betmdl = beta_df.loc[:, f"BET{plane}{MDL}"].to_numpy()
    alfmdl = beta_df.loc[:, f"ALF{plane}{MDL}"].to_numpy()
    with np.errstate(divide='ignore'):
        cot_phase_meas = 1 / np.tan(tilted_meas)
        cot_phase_model = 1 / np.tan(tilted_model)
    # calculate enumerators and denominators for far more cases than needed
    # shift1 are the cases BBA, ABB, AxBB, AxxBB etc. (the used BPMs are adjacent)
    # shift2 are the cases where the used BPMs are separated by one. only BAB is used for  3-BPM
    cot_phase_meas_shift1 = cot_phase_meas - np.roll(cot_phase_meas, -1, axis=0)
    cot_phase_model_shift1 = cot_phase_model - np.roll(cot_phase_model, -1, axis=0) + EPSILON
    cot_phase_meas_shift2 = cot_phase_meas - np.roll(cot_phase_meas, -2, axis=0)
    cot_phase_model_shift2 = cot_phase_model - np.roll(cot_phase_model, -2, axis=0) + EPSILON
    # calculate the sum of the fractions
    bet_frac = (cot_phase_meas_shift1[0]/cot_phase_model_shift1[0] +
                cot_phase_meas_shift1[3]/cot_phase_model_shift1[3] +
                cot_phase_meas_shift2[1]/cot_phase_model_shift2[1]) / 3

    alf_mdl_term = (((cot_phase_model + np.roll(cot_phase_model, -1, axis=0))[0] + 2.0 * alfmdl)
                + ((cot_phase_model + np.roll(cot_phase_model, -1, axis=0))[3] + 2.0 * alfmdl)
                + ((cot_phase_model + np.roll(cot_phase_model, -2, axis=0))[1] + 2.0 * alfmdl)) / 6.0
    alf_meas_term = (((cot_phase_meas + np.roll(cot_phase_meas, -1, axis=0))[0])
                + ((cot_phase_meas + np.roll(cot_phase_meas, -1, axis=0))[3])
                + ((cot_phase_meas + np.roll(cot_phase_meas, -2, axis=0))[1])) / 6.0

    # multiply the fractions by betmdl and calculate the arithmetic mean
    beti = bet_frac * betmdl
    alfi = bet_frac * alf_mdl_term + alf_meas_term
    # calculate errphi_ij^2 / sin^2 phimdl_ij * beta
    with np.errstate(divide='ignore', invalid='ignore'):
        sin_squared_model = tilted_errmeas**2 / np.square(np.sin(tilted_model)) * betmdl
        sin_quadrup_model = tilted_errmeas**2 / np.power(np.sin(tilted_model), 4) * betmdl
    # square it again beacause it's used in a vector length
    sin_squared_model = np.square(sin_squared_model)
    sin_squ_model_shift1 = sin_squared_model + np.roll(sin_squared_model, -1, axis=0) / np.square(cot_phase_model_shift1)
    sin_squ_model_shift2 = sin_squared_model + np.roll(sin_squared_model, -2, axis=0) / np.square(cot_phase_model_shift2)
    sin_quad_model_shift1 = sin_quadrup_model + np.roll(sin_quadrup_model, -1, axis=0) / np.square(cot_phase_model_shift1)
    sin_quad_model_shift2 = sin_quadrup_model + np.roll(sin_quadrup_model, -2, axis=0) / np.square(cot_phase_model_shift2)
    beterr = np.sqrt(sin_squ_model_shift1[0] + sin_squ_model_shift1[3] + sin_squ_model_shift2[1]) / 3
    beta_df["BET" + plane] = beti
    beta_df["ERRBET" + plane] = beterr
    beta_df["ALF" + plane] = alfi
    beta_df["ERRALF" + plane] = alf_mdl_term * beterr / betmdl + 0.5 * np.sqrt(
        sin_quad_model_shift1[0] + sin_quad_model_shift1[3] + sin_quad_model_shift2[1])
    beta_df = _get_delta_columns(beta_df, plane)
    return beta_df


def _get_delta_columns(beta_df, plane):
    beta_df[f"{DELTA}BET{plane}"] = df_rel_diff(beta_df, f"BET{plane}", f"BET{plane}{MDL}")
    beta_df[f"{ERR}{DELTA}BET{plane}"] = df_ratio(beta_df, f"{ERR}BET{plane}", f"BET{plane}{MDL}")
    beta_df[f"{DELTA}ALF{plane}"] = df_diff(beta_df, f"ALF{plane}", f"ALF{plane}{MDL}")
    beta_df[f"{ERR}{DELTA}ALF{plane}"] = beta_df.loc[:, f"{ERR}ALF{plane}"].to_numpy()
    return beta_df


def _tilt_slice_matrix(matrix, slice_shift, slice_width, tune=0):
    """
    Tilts and slices the ``matrix``.
    Tilting means shifting each column upwards one step more than the previous columnns, i.e.

    a a a a a       a b c d
    b b b b b       b c d e
    c c c c c  -->  c d e f
    ...             ...
    y y y y y       y z a b
    z z z z z       z a b c
    """
    invrange = matrix.shape[0] - 1 - np.arange(matrix.shape[0])
    matrix[matrix.shape[0] - slice_shift:, :slice_shift] += tune
    matrix[:slice_shift, matrix.shape[1] - slice_shift:] -= tune
    return np.roll(matrix[np.arange(matrix.shape[0]), circulant(invrange)[invrange]],
                   slice_shift, axis=0)[:slice_width]


def _get_filtered_model_df(meas_input, phase, plane, best=False):
    model = _try_best_model(meas_input) if best else meas_input.accelerator.model
    df = pd.DataFrame(model).loc[phase["MEAS"].index, ["S", f"BET{plane}", f"ALF{plane}", f"MU{plane}"]]
    if not best:
        df.rename(columns={f"BET{plane}": f"BET{plane}{MDL}", f"ALF{plane}": f"ALF{plane}{MDL}", f"MU{plane}": f"MU{plane}{MDL}"}, inplace=True)
    return df


def _try_best_model(meas_input):
    if meas_input.accelerator.model_best_knowledge is None:
        LOGGER.debug("No best knowledge model - using the normal one.")
        return meas_input.accelerator.model
    return meas_input.accelerator.model_best_knowledge



"""
Segment Creator
---------------

This module provides convenience functions for model creation of a ``segment``.
"""
import shutil
from pathlib import Path

from omc3.model.constants import MACROS_DIR, GENERAL_MACROS
from omc3.utils import logging_tools
from omc3.utils.iotools import create_dirs

LOGGER = logging_tools.get_logger(__name__)


class SegmentCreator(object):
    @classmethod
    def prepare_run(cls, accel):
        macros_path = accel.model_dir / MACROS_DIR
        create_dirs(macros_path)
        lib_path = Path(__file__).parent.parent.parent / "lib"
        shutil.copy(lib_path / GENERAL_MACROS, macros_path / GENERAL_MACROS)

    @classmethod
    def get_madx_script(cls, accel):
        madx_template = accel.get_file("segment.madx").read_text()
        replace_dict = {
            "MAIN_SEQ": accel.load_main_seq_madx(),  # LHC only
            "OPTICS_PATH": accel.modifiers,  # all
            "NUM_BEAM": accel.beam,  # LHC only
            "PATH": accel.model_dir,  # all
            "OUTPUT": accel.model_dir,  # Booster only
            "LABEL": accel.label,  # all
            "BETAKIND": accel.kind,  # all
            "STARTFROM": accel.start.name,  # all
            "ENDAT": accel.end.name,  # all
            "RING": accel.ring,  # Booster only
            "KINETICENERGY": accel.energy,  # PS only
            "FILES_DIR": accel.get_dir(),  # Booster and PS
            "NAT_TUNE_X": accel.nat_tunes[0],  # Booster and PS
            "NAT_TUNE_Y": accel.nat_tunes[1],  # Booster and PS
        }
        return madx_template % replace_dict



"""
Dpp
---

This module contains deltap over p calculations related functionality of ``optics_measurements``.
It provides functions to computes and arrange dp over p.
"""
import logging

import numpy as np
import pandas as pd

DPP_TOLERANCE = 1e-4
AMP_DPP_TOLERANCE = 1e-5
LOGGER = logging.getLogger(__name__)


def arrange_dpps(dpps):
    """
    Grouping of dpp-values and averaging them in the bins, also zeroes the bin closest to zero.
    """
    closest_to_zero = np.argmin(np.abs(dpps))
    ranges = _compute_ranges(dpps)
    zero_offset = np.mean(_values_in_range(_find_range_with_element(ranges, closest_to_zero), dpps))
    LOGGER.debug(f"dp/p closest to zero is {dpps[closest_to_zero]}")
    if np.abs(zero_offset) > DPP_TOLERANCE:
        LOGGER.warning(f"Analysed files have large momentum deviation {zero_offset}. "
                       f"Optics parameters might be wrong.")
    LOGGER.debug(f"Detected dpp differences, aranging as: {ranges}, zero offset: {zero_offset}.")
    arranged_dpps = []
    for idx in range(len(dpps)):
        dpps_from_range = _values_in_range(_find_range_with_element(ranges, idx), dpps)
        arranged_dpps.append(np.mean(dpps_from_range) - zero_offset)
    return np.array(arranged_dpps)


def _values_in_range(range_to_use, dpp_values):
    return [dpp_values[idx] for idx in range_to_use]


def _find_range_with_element(ranges, element):
    return [dpp_range for dpp_range in ranges if element in dpp_range][0]


def _compute_ranges(dpps):
    ordered_ids = np.argsort(dpps)
    ranges = [[ordered_ids[0]]]
    if len(ordered_ids) == 1:
        return ranges
    for idx in ordered_ids[1:]:
        if abs(dpps[ranges[-1][0]] - dpps[idx]) < DPP_TOLERANCE:
            ranges[-1].append(idx)
        else:
            ranges.append([idx])
    return ranges


def append_amp_dpp(list_of_tfs, dpp_values):
    for i, dpp in enumerate(dpp_values):
        list_of_tfs[i].headers["DPPAMP"] = dpp if (not np.isnan(dpp) and np.abs(dpp) > AMP_DPP_TOLERANCE) else 0.0
    return list_of_tfs


def append_dpp(list_of_tfs, dpp_values):
    for i, dpp in enumerate(dpp_values):
        list_of_tfs[i].headers["DPP"] = dpp
    return list_of_tfs


def calculate_dpoverp(input_files, meas_input): 
    df_orbit = pd.DataFrame(meas_input.accelerator.model).loc[:, ['S', 'DX']]
    df_orbit = pd.merge(df_orbit, input_files.joined_frame('X', ['CO', 'CORMS']), how='inner',
                        left_index=True, right_index=True)
    mask = meas_input.accelerator.get_element_types_mask(df_orbit.index, ["arc_bpm"])
    df_filtered = df_orbit.loc[mask, :]
    dispersions = df_filtered.loc[:, "DX"].to_numpy()
    denom = np.sum(dispersions ** 2)
    if denom == 0.:
        raise ValueError("Cannot compute dpp probably no arc BPMs.")
    amps = input_files.get_data(df_filtered, "CO")
    if amps.ndim == 1:
        return np.sum(dispersions * amps) / denom
    else:
        numer = np.sum(dispersions[:, None] * input_files.get_data(df_filtered, "CO"), axis=0)
    return numer / denom


def calculate_amp_dpoverp(input_files, meas_input):
    df_orbit = pd.DataFrame(meas_input.accelerator.model).loc[:, ['S', 'DX']]
    df_orbit = pd.merge(df_orbit, input_files.joined_frame('X', ['AMPX', 'AMPZ']), how='inner',
                        left_index=True, right_index=True)
    # TODO often missing AMPZ causes FutureWarning in future will be KeyError
    if np.all(np.isnan(input_files.get_data(df_orbit, "AMPZ"))):
        return np.zeros(len(input_files["X"]))
    mask = meas_input.accelerator.get_element_types_mask(df_orbit.index, ["arc_bpm"])
    df_filtered = df_orbit.loc[mask, :]
    amps = input_files.get_data(df_filtered, 'AMPX') * input_files.get_data(df_filtered, 'AMPZ')
    mask_zeros = (amps > 0) if amps.ndim == 1 else (np.sum(amps, axis=1) > 0)
    dispersions = df_filtered.loc[mask_zeros, "DX"]
    denom = np.sum(dispersions ** 2)
    if denom == 0.:
        raise ValueError("Cannot compute dpp probably no arc BPMs.")
    if amps.ndim == 1:
        numer = np.sum(dispersions * amps[mask_zeros])
    else:
        numer = np.sum(dispersions[:, None] * amps[mask_zeros, :], axis=0)
    return numer / denom



r"""
Outliers
--------

Helper functions for outlier detection.
"""
import numpy as np
from numpy.typing import ArrayLike
from scipy.stats import t


def get_filter_mask(data: ArrayLike, x_data: ArrayLike = None, limit: float = 0.0,
                    niter: int = 20, nsig: int = None, mask: ArrayLike = None) -> ArrayLike:
    """
    Filters the array of values which are meant to be constant or a linear function of the x-data
    array if that is provided, by checking how many sigmas they are deviating from the average.
    Returns a filter mask for the original array (meaning ``True`` for elements that should be kept).

    Args:
         data (ArrayLike): Data to filter.
         x_data (ArrayLike): Additional x-data. If given a linear function is fit on the data,
                             and the outliers filtered on the difference to this function.
                             Otherwise, the distribution of data is used directly for filtering.
         limit (float): Only data deviating more than this limit from the average is filtered.
         niter (int): Maximum number of filter iterations to do.
                      Iterations are interrupted if the number of datapoints did not shrink
                      in the last iteration or if there is only two or less data-points left.
         nsig (int): number of sigmas within the data points are considered okay, outside they are
                     considered outliers. If not given, nsigmas is got from the
                     1-0.5/length(data) percentile of a student-t distribution.
         mask (ArrayLike): Boolean mask of data-points to consider.
                           If ``None``, all data is considered.

    Returns:
        ArrayLike: Boolean array containing ``True`` entries for data-points that are good,
                   and ``False`` for the ones that should be filtered.

    """
    if x_data is not None:
        if not len(data) == len(x_data):
            raise ValueError("Datasets are not equally long.")
    
    if mask is not None:
        if not len(data) == len(mask):
            raise ValueError("Mask is not equally long as dataset.")
    else:
        mask = np.ones_like(data, dtype=bool)

    if nsig is None:
        nsig = _get_significance_cut_from_length(np.sum(mask))

    # To fulfill the condition for the first iteration:
    prevlen = np.sum(mask) + 1

    # Cleaning iteration
    for _ in range(niter):
        # check that number of points decreases and some are left
        if not ((np.sum(mask) < prevlen) and
                (np.sum(mask) > 2)):
            break
        prevlen = np.sum(mask)

        # get the (remaining) data-points used to get moments from,
        # if x-data is given, fit and remove slope.
        if x_data is None:
            y, y_orig = _get_data(mask, data)
        else:
            y, y_orig = _get_data_without_slope(mask, x_data, data)

        # keep only points within nsig sigma of the remaining data.
        avg, std = np.mean(y), np.std(y)
        mask = np.logical_and(mask, np.abs(y_orig - avg) < np.max([limit, nsig * std]))
    return mask


def _get_data_without_slope(mask, x, y):
    m, b = np.polyfit(x[mask], y[mask], 1)
    return y[mask] - b - m * x[mask], y - b - m * x


def _get_data(mask, data):
    return data[mask], data


# Set the sigma cut, that expects 1 value to be cut
# if it is sample of normal distribution
def _get_significance_cut_from_length(length):
    return t.ppf(1 - 0.5 / length, length)



"""
Coupling
--------

This module contains linear coupling calculations related functionality of ``optics_measurements``.
It provides functions to computes and the coupling resonance driving terms, which are part of the standard
optics outputs.
"""
from collections import OrderedDict
from functools import partial, reduce
from pathlib import Path
from typing import Callable, Dict, List, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import tfs
from numpy import cos, exp, ndarray, sin, sqrt, tan
from optics_functions.coupling import coupling_via_cmatrix

from omc3.definitions.constants import PI2, PI2I
from omc3.harpy.constants import COL_MU
from omc3.optics_measurements.beta_from_phase import _tilt_slice_matrix
from omc3.optics_measurements.constants import (
    AMPLITUDE,
    F1001,
    F1010,
    IMAG,
    REAL,
    NAME,
    PHASE,
    PHASE_ADV,
    SECONDARY_AMPLITUDE_X,
    SECONDARY_AMPLITUDE_Y,
    SECONDARY_FREQUENCY_X,
    SECONDARY_FREQUENCY_Y,
    S,
    ERR,
    EXT,
    MDL,
    DELTA
)
from omc3.utils import logging_tools, stats

LOGGER = logging_tools.get_logger(__name__)

COLS_TO_KEEP_X: List[str] = [NAME, S, f"{AMPLITUDE}01", f"{PHASE}01", f"{PHASE_ADV}X"]
COLS_TO_KEEP_Y: List[str] = [NAME, S, f"{AMPLITUDE}10", f"{PHASE}10", f"{PHASE_ADV}Y"]
CUTOFF: int = 5


def calculate_coupling(
    meas_input: dict,
    input_files: dict,
    phase_dict: Dict[str, Tuple[Dict[str, tfs.TfsDataFrame], Sequence[tfs.TfsDataFrame]]],
    tune_dict: Dict[str, float],
    header_dict: OrderedDict,
) -> None:
    """
    Calculates the coupling RDTs f1001 and f1010, as well as the closest tune approach Cminus (|C-|).
    This represents the "2 BPM method" in https://cds.cern.ch/record/1264111/files/CERN-BE-Note-2010-016.pdf
    (a more up-to-date reference will come in the near future).

    Two formulae are used to calculate the Cminus, taken from the following reference:
    https://cds.cern.ch/record/2135848/files/PhysRevSTAB.17.051004.pdf
    The first one (Eq(1)) is an approximation using only the amplitudes of the RDTs, while the second one
    (Eq(2) in the same paper) is more exact but needs also the phase of the RDT.

    The results are written down in the optics_measurements outputs as **f1001.tfs** and **f1010.tfs** files.

    Args:
        meas_input (dict): `OpticsInput` object containing analysis settings from the command-line.
        input_files (dict): `InputFiles` (dict) object containing frequency spectra files (linx/y) for
            each transverse plane (as keys).
        phase_dict (Dict[str, Tuple[Dict[str, tfs.TfsDataFrame], tfs.TfsDataFrame]]): dictionary containing
            the measured phase advances, with an entry for each transverse plane. In said entry is a
            dictionary with the measured phase advances for 'free' and 'uncompensated' cases, as well as
            the location of the output ``TfsDataFrames`` for the phases.
        tune_dict (Dict[str, float]): `TuneDict` object containing measured tunes. There is an entry
            calculated for the 'Q', 'QF', 'QM', 'QFM' and 'ac2bpm' modes, each value being a float.
        header_dict (OrderedDict): header dictionary of common items for coupling output files,
            will be attached as the header to the **f1001.tfs** and **f1010.tfs** files..
    """
    LOGGER.info("Calculating coupling")
    bd = meas_input.accelerator.beam_direction
    compensation = "uncompensated" if meas_input.compensation == "model" else "free"

    # We need vertical and horizontal spectra, so we have to intersect first all inputs with X and Y phase
    # output furthermore the output has to be rearranged in the order of the model (important for e.g. LHC
    # beam 2) and thus we need to intersect the *model* index with all the above. Since the first index of
    # the .intersect chain dictates the order, we have to start with the model index.
    LOGGER.debug("Intersecting measurements, starting with model")
    joined: tfs.TfsDataFrame = _joined_frames(input_files)  # merge transverse input frames
    joined_index: pd.Index = (
        meas_input.accelerator.model.index.intersection(joined.index)
        .intersection(phase_dict["X"][compensation]["MEAS"].index)
        .intersection(phase_dict["Y"][compensation]["MEAS"].index)
    )
    joined = joined.loc[joined_index].copy()

    phases_x: tfs.TfsDataFrame = phase_dict["X"][compensation]["MEAS"].loc[joined_index, joined_index].copy()
    phases_y: tfs.TfsDataFrame = phase_dict["Y"][compensation]["MEAS"].loc[joined_index, joined_index].copy()

    LOGGER.debug("Averaging (arithmetic mean) amplitude columns")
    for col in [SECONDARY_AMPLITUDE_X, SECONDARY_AMPLITUDE_Y]:
        arithmetically_averaved_columns = [c for c in joined.columns if c.startswith(col)]
        joined[col] = stats.weighted_mean(joined[arithmetically_averaved_columns], axis=1)

    LOGGER.debug("Averaging (circular mean) frequency columns")  # make sure to use period=1 here
    for col in [SECONDARY_FREQUENCY_X, SECONDARY_FREQUENCY_Y]:
        circularly_averaved_columns = [x for x in joined.columns if x.startswith(col)]
        joined[col] = bd * stats.circular_mean(
            joined[circularly_averaved_columns], axis=1, period=1
        )

    LOGGER.debug("Finding BPM pairs for momentum reconstruction")
    bpm_pairs_x, deltas_x = _find_pair(phases_x, meas_input.coupling_pairing)
    bpm_pairs_y, deltas_y = _find_pair(phases_y, meas_input.coupling_pairing)

    LOGGER.debug("Computing complex lines from spectra")
    A01: np.ndarray = 0.5 * _get_complex_line(
        joined[SECONDARY_AMPLITUDE_X] * exp(joined[SECONDARY_FREQUENCY_X] * PI2I), deltas_x, bpm_pairs_x
    )
    B10: np.ndarray = 0.5 * _get_complex_line(
        joined[SECONDARY_AMPLITUDE_Y] * exp(joined[SECONDARY_FREQUENCY_Y] * PI2I), deltas_y, bpm_pairs_y
    )
    A0_1: np.ndarray = 0.5 * _get_complex_line(
        joined[SECONDARY_AMPLITUDE_X] * exp(-joined[SECONDARY_FREQUENCY_X] * PI2I), deltas_x, bpm_pairs_x
    )
    B_10: np.ndarray = 0.5 * _get_complex_line(
        joined[SECONDARY_AMPLITUDE_Y] * exp(-joined[SECONDARY_FREQUENCY_Y] * PI2I), deltas_y, bpm_pairs_y
    )

    q1001_from_A = -np.angle(A01) + (bd * joined[f"{COL_MU}Y"].to_numpy() - 0.25) * PI2
    q1001_from_B = np.angle(B10) - (bd * joined[f"{COL_MU}X"].to_numpy() - 0.25) * PI2
    eq_1001 = exp(1.0j * q1001_from_A) + exp(1.0j * q1001_from_B)

    q1010_from_A = -np.angle(A0_1) - (bd * joined[f"{COL_MU}Y"].to_numpy() - 0.25) * PI2
    q1010_from_B = -np.angle(B_10) - (bd * joined[f"{COL_MU}X"].to_numpy() - 0.25) * PI2
    eq_1010 = exp(1.0j * q1010_from_A) + exp(1.0j * q1010_from_B)

    LOGGER.debug("Computing average of coupling RDTs")
    f1001: np.ndarray = -0.5 * sqrt(np.abs(A01 * B10)) * eq_1001 / abs(eq_1001)
    f1010: np.ndarray = 0.5 * sqrt(np.abs(A0_1 * B_10)) * eq_1010 / abs(eq_1010)

    LOGGER.debug("Getting tune separation from measurements")
    tune_separation = np.abs(tune_dict["X"]["QFM"] % 1.0 - tune_dict["Y"]["QFM"] % 1.0)

    LOGGER.debug("Calculating approximated Cminus")
    C_approx = 4.0 * tune_separation * np.mean(np.abs(f1001))
    header_dict["Cminus_approx"] = C_approx
    LOGGER.info(
        f"|C-| (approx) = {C_approx:.5f}, tune_sep = {tune_separation:.3f}, from Eq.1 in PRSTAB 17,051004"
    )

    LOGGER.debug("Calculating exact Cminus")
    C_exact = np.abs(4.0 * tune_separation * np.mean(f1001 * exp(1.0j * (joined[f"{COL_MU}X"] - joined[f"{COL_MU}Y"]))))
    header_dict["Cminus_exact"] = C_exact
    LOGGER.info(f"|C-| (exact)  = {C_exact:.5f}, from Eq.2 w/o i*s*Delta/R in PRSTAB 17,051004")

    if meas_input.compensation == "model":
        LOGGER.debug("Compensating coupling RDT values by model")
        f1001, f1010 = compensate_rdts_by_model(f1001, f1010, tune_dict)

    LOGGER.debug("Adding model values and deltas")
    model_coupling = coupling_via_cmatrix(meas_input.accelerator.model).loc[joined_index]

    f1001_df = _rdt_to_output_df(f1001, model_coupling[F1001], meas_input.accelerator.model, joined_index)
    f1010_df = _rdt_to_output_df(f1010, model_coupling[F1010], meas_input.accelerator.model, joined_index)

    tfs.write(Path(meas_input.outputdir) / f"{F1001.lower()}{EXT}", f1001_df, header_dict)
    tfs.write(Path(meas_input.outputdir) / f"{F1010.lower()}{EXT}", f1010_df, header_dict)


def compensate_rdts_by_model(
    f1001: np.ndarray, f1010: np.ndarray, tune_dict: Dict[str, float]
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compensate coupling RDTs by model (equation) only, implies we're providing a driven model (ACD / ATD
    kick). The scaling factors are calculated from the model's free and driven tunes, and the scaled RDTs
    are returned.

    Args:
        f1001 (np.ndarray): the pre-calculated driven coupling RDTs as an array.
        f1010 (np.ndarray): the pre-calculated driven coupling RDTs as an array.
        tune_dict (Dict[str, float]): `TuneDict` object containing measured tunes. There is an entry
            calculated for the 'Q', 'QF', 'QM', 'QFM' and 'ac2bpm' modes, each value being a float.

    Returns:
        The scaled RDTs.
    """
    f1001 = np.array(f1001)  # make sure we don't modify inplace
    f1010 = np.array(f1010)

    LOGGER.debug("Retrieving model's natural tunes")  # QFM is free since model is ACD/ADT- driven model
    Qx = PI2 * tune_dict["X"]["QFM"]
    Qy = PI2 * tune_dict["Y"]["QFM"]

    LOGGER.debug("Retrieving model's driven tunes")  # QM is driven as model is ACD/ADT-driven model
    Qx_driven = PI2 * tune_dict["X"]["QM"]
    Qy_driven = PI2 * tune_dict["Y"]["QM"]

    LOGGER.debug("Computing scaling factor from driven model")
    f1001_scaling_factor = np.sqrt(np.abs(sin(Qy_driven - Qx) * sin(Qx_driven - Qy))) / np.abs(sin(Qx - Qy))
    f1010_scaling_factor = np.abs(np.sqrt(sin(Qx + Qy_driven) * sin(Qy + Qx_driven)) / sin(Qx + Qy))
    f1001 *= f1001_scaling_factor
    f1010 *= f1010_scaling_factor

    LOGGER.info("Compensation by model:")
    LOGGER.info(f"\t f1001 scaling factor: {f1001_scaling_factor:.5f}")
    LOGGER.info(f"\t f1010 scaling factor: {f1010_scaling_factor:.5f}")
    return f1001, f1010


def compensate_rdts_ryoichi():
    pass


# ----- Helpers ----- #

def _find_pair(phases: tfs.TfsDataFrame, mode: int = 1):
    """
    Does the BPM pairing for coupling calculation.

    Args:
        mode (int): Value to determine the BPM pairing. If ``0`` is given,
            tries to find the best candidate. If a value ``n>=1`` is given,
            then takes the n-th following BPM downstream for the pairing.
    """

    if mode == 0:
        return _find_candidate(phases)
    else:
        return _take_next(phases, mode)


def _take_next(phases: tfs.TfsDataFrame, shift: int = 1):
    """
    Takes the following BPM for momentum reconstruction by a given shift.

    Args:
        phases (tfs.TfsDataFrame): Dataframe matrix of phase advances, as calculated in phase.py.
        shift (int): Value to determine the BPM pairing. If ``0`` is given,
           tries to find the best candidate. If a value ``n>=1`` is given,
           then takes the n-th following BPM downstream for the pairing.
   """
    indices = np.roll(np.arange(phases.to_numpy().shape[0]), shift)
    return indices, phases.to_numpy()[np.arange(phases.to_numpy().shape[0]), indices] - 0.25


def _find_candidate(phases: tfs.TfsDataFrame) -> Tuple[np.ndarray, np.ndarray]:
    """
    Finds the best candidate for momentum reconstruction.

    Args:
      phases (tfs.TfsDataFrame): Dataframe matrix of phase advances, as calculated in phase.py.
      bd (int): beam direction, will be negative for beam 2.

    Returns:
        The indices of best candidates, and the corresponding phase advances between these indices.
    """
    slice_ = _tilt_slice_matrix(phases.to_numpy(), 0, 2 * CUTOFF) - 0.25  # do not overwrite built-in 'slice'
    indices = np.argmin(abs(slice_), axis=0)
    deltas = slice_[indices, range(len(indices))]
    indices = (indices + np.arange(len(indices))) % len(indices)
    return np.array(indices), deltas


def _get_complex_line(
    spectral_lines: Union[pd.Series, np.ndarray],
    deltas: Union[pd.Series, np.ndarray],
    pairs: Union[pd.Series, np.ndarray],
) -> np.ndarray:
    """
    Calculates the complex line from the real lines at positions i and j, where j is determined by
    taking the next BPM with a phase advance sufficiently close to pi/2

    Args:
      spectral_lines (Union[pd.Series, np.ndarray]): vector with measured (real) spectral lines.
      deltas (Union[pd.Series, np.ndarray]): vector with phase advances minus 90deg.
      pairs (Union[pd.Series, np.ndarray]): vector with indices for pairing.

    Returns:
        A numpy array with the results at all given positions from the inputs.
    """
    spectral_lines = np.array(spectral_lines)  # make sure we avoid any inplace modification of data
    deltas = np.array(deltas)
    pairs = np.array(pairs)
    return (1.0 - 1.0j * tan(PI2 * deltas)) * spectral_lines - 1.0j / cos(PI2 * deltas) * spectral_lines[
        pairs
    ]


def _rdt_to_output_df(
    fterm: Union[pd.Series, np.ndarray],
    fterm_mdl: Union[pd.Series, np.ndarray],
    model_df: tfs.TfsDataFrame,
    index: pd.Index,
) -> pd.DataFrame:
    """
    Creates the output coupling RDT dataframe from the given RDT and its model-calculated counterpart.
    Combines all the needed columns (``S``, ``NAME``, ``AMP``, ``PHASE``, ``REAL``, ``IMAG``, and the
    ``ERR*`` and ``*MDL`` columns).

    .. note::
        At the moment, the dataframe holds a ``DELTAREAL`` and ``DELTAIMAG`` columns, which are calculated
        as `REAL - REALMDL` and `IMAG - IMAGMDL`. As most of the time the model is coupling-free, these are
        usually going to be identical values to ``REAL`` and ``IMAG`` columns. They are still included in
        case one wishes to have a coupled model, as some machines sometimes do (but not LHC afaik).

        Similarly, there are ``ERRDELTAREAL`` and ``ERRDELTAIMAG`` columns, which are at the moment
        the same values as ``ERRREAL`` and ``ERRIMAG`` columns. In the future, we might want to have
        a fancier calculation for these.

    .. important::
        The columns mentionned in the note above are required and expected in the correction calculation.
        It would fail the correction functionality to remove these columns.

    Args:
        fterm (Union[pd.Series, np.ndarray]): the calculated coupling RDT.
        fterm_mdl (Union[pd.Series, np.ndarray]): corresponding RDT values calculated from the model (e.g.
            calculated via cmatrix from the model_df).
        model_df (tfs.TfsDataFrame): the model dataframe attached to the accelerator object, used to get the
            ``S`` position.
        index (pd.Index): the joined intersected index used to align everything.

    Returns:
        pd.DataFrame: dataframe ready to be written to the out file `.tfs`
    """
    df = pd.DataFrame()
    df[S] = model_df.loc[index, S]
    df[NAME] = index

    LOGGER.debug("Computing RDT amplitude values")
    df[AMPLITUDE] = np.abs(fterm)
    df[AMPLITUDE + MDL] = np.abs(fterm_mdl)

    LOGGER.debug("Computing phase values")
    df[PHASE] = np.angle(fterm)
    df[PHASE + MDL] = np.angle(fterm_mdl)

    LOGGER.debug("Computing deviation from model")
    df[DELTA + AMPLITUDE] = df[AMPLITUDE] - df[AMPLITUDE + MDL]
    df[DELTA + PHASE] = df[PHASE] - df[PHASE + MDL]

    LOGGER.debug("Computing error values")
    df[ERR + AMPLITUDE] = 0  # TODO: will need to implement this calculation later
    df[ERR + PHASE] = 0

    LOGGER.debug("Adding real and imaginary parts columns")
    df[REAL] = np.real(fterm)
    df[REAL + MDL] = np.real(fterm_mdl)
    df[ERR + REAL] = 0  # TODO: same
    # These following columns are needed in the correction calculation later on
    # Most of the time model has 0 coupling so the DELTA is just the REAL / IMAG but let's
    # not neglect that we might want to have weird coupled models sometimes
    # For now error on delta is just the error on REAL / IMAG but in the future
    # we might want to change this for a fancier calculation
    df[DELTA + REAL] = df[REAL] - df[REAL + MDL]
    df[ERR + DELTA + REAL] = df[ERR + REAL]

    df[IMAG] = np.imag(fterm)
    df[IMAG + MDL] = np.imag(fterm_mdl)
    df[ERR + IMAG] = 0  # TODO: same
    # See comment above, same thing here for IMAG
    df[DELTA + IMAG] = df[IMAG] - df[IMAG + MDL]
    df[ERR + DELTA + IMAG] = df[ERR + IMAG]

    return df.sort_values(by=S)


def _joined_frames(input_files: dict) -> tfs.TfsDataFrame:
    """
    Merges spectrum data from the two planes from all the input files.

    Args:
        input_files (dict): `InputFiles` (dict) object containing frequency spectra files (linx/y) for each
            transverse plane (as keys).
    """
    joined_dfs = []
    assert len(input_files["X"]) == len(input_files["Y"])

    for i, (linx, liny) in enumerate(zip(input_files["X"], input_files["Y"])):
        linx = linx[COLS_TO_KEEP_X].copy()
        liny = liny[COLS_TO_KEEP_Y].copy()

        linx = linx.rename(columns=rename_col("X", i))
        liny = liny.rename(columns=rename_col("Y", i))

        merged_transverse_lins_df = pd.merge(
            left=linx,
            right=liny,
            on=[NAME],
            how="inner",
            sort=False,
        )
        joined_dfs.append(merged_transverse_lins_df)

    partial_merge = partial(pd.merge, how="inner", on=[NAME], sort=False)
    reduced = reduce(partial_merge, joined_dfs).set_index(NAME)
    reduced = reduced.rename(
        columns={f"{PHASE_ADV}X_X_0": f"{PHASE_ADV}X", f"{PHASE_ADV}Y_Y_0": f"{PHASE_ADV}Y"}
    )
    return tfs.TfsDataFrame(reduced)


def rename_col(plane: str, index: int) -> Callable:
    """
    Generate appropriate column name for renaming before merging dataframes from InputFiles.

    Args:
        plane (str): plane for which to apply the renaming.
        index (int): index location of the df which columns are renamed in the input files.

    Returns:
        The renaming function callable to be given to pandas.
    """

    def fn(column):
        if column == NAME:
            return column
        return f"{column}_{plane}_{index}"

    return fn



"""
ESRF
----

Accelerator-Class for the ``ESRF`` machine.

Model Creation Keyword Args:
    *--Optional--*

    - **dpp** *(float)*:

        Deltap/p to use.

        default: ``0.0``


    - **driven_excitation** *(str)*:

        Denotes driven excitation by `AC-dipole` (acd) or by `ADT` (adt)

        choices: ``('acd', 'adt')``


    - **drv_tunes** *(float)*:

        Driven tunes without integer part.


    - **energy** *(float)*:

        Energy in **Tev**.


    - **model_dir** *(str)*:

        Path to model directory; loads tunes and excitation from model!


    - **modifiers** *(str)*:

        Path to the optics file to use (modifiers file).


    - **nat_tunes** *(float)*:

        Natural tunes without integer part.


    - **xing**:

        If True, x-ing angles will be applied to model

        action: ``store_true``
"""
from generic_parser import EntryPoint

from omc3.model.accelerators.accelerator import AccElementTypes, Accelerator


class Esrf(Accelerator):
    NAME = "esrf"
    RE_DICT = {AccElementTypes.BPMS: r"BPM",
               AccElementTypes.MAGNETS: r".*",
               AccElementTypes.ARC_BPMS: r"BPM\.(\d*[02468]\.[1-5]|\d*[13579]\.[3-7])",
               }  # bpms 1-5 in even cells and bpms 3-7 in odd cells.

    def __init__(self, *args, **kwargs):
        parser = EntryPoint(self.get_parameters(), strict=True)
        opt = parser.parse(*args, **kwargs)
        super().__init__(opt)



"""
Constants
---------

General constants to use throughout ``omc3``, so they don't have to be redefined all the time.
Also helps with consistency.
"""
from typing import Dict, Tuple

import numpy as np

PI: float = np.pi
PI2: float = 2 * np.pi
PI2I: float = 2j * np.pi
PLANES: Tuple[str, str] = ("X", "Y")
PLANE_TO_NUM: Dict[str, int] = dict(X=1, Y=2)
PLANE_TO_HV: Dict[str, str] = dict(X="H", Y="V")

UNIT_IN_METERS: Dict[str, float] = dict(
    km=1e3, m=1e0, mm=1e-3, um=1e-6, nm=1e-9, pm=1e-12, fm=1e-15, am=1e-18
)



"""
Manager
-------

This module provides high-level functions to manage most functionality of ``model``.
It contains entrypoint wrappers to get accelerator classes or their instances.
"""
from generic_parser.entrypoint_parser import entrypoint, EntryPoint, EntryPointParameters
from omc3.model.accelerators import lhc, ps, esrf, psbooster, skekb, petra, iota

ACCELS = {
    lhc.Lhc.NAME: lhc.Lhc,
    ps.Ps.NAME: ps.Ps,
    esrf.Esrf.NAME: esrf.Esrf,
    psbooster.Psbooster.NAME: psbooster.Psbooster,
    skekb.SKekB.NAME: skekb.SKekB,
    "JPARC": skekb.SKekB,
    petra.Petra.NAME: petra.Petra,
    iota.Iota.NAME: iota.Iota
}


def _get_params():
    params = EntryPointParameters()
    params.add_parameter(name="accel", required=True, choices=list(ACCELS.keys()),
                         help="Choose the accelerator to use.Can be the class already.")
    return params


@entrypoint(_get_params())
def get_accelerator(opt, other_opt):
    """
    Returns the `Accelerator` instance of the desired accelerator, as given at the commandline.
    """
    if not isinstance(opt.accel, str):
        # assume it's the class
        return opt.accel
    return ACCELS[opt.accel](other_opt)


@entrypoint(_get_params())
def get_parsed_opt(opt, other_opt):
    """Get all accelerator related options as a `dict`."""
    accel = ACCELS[opt.accel]
    parser = EntryPoint(accel.get_parameters(), strict=True)
    accel_opt = parser.parse(other_opt)
    return {**opt, **accel_opt}



"""
Frequency
---------

This module contains the frequency calculations related functionality of ``harpy``.

It provides functions to calculate the frequency spectra of turn-by-turn data.
This calculation is done using a combination of SVD decomposition zero_padded `fft` to speed up the
analysis.
Also searches for resonances in the calculated spectra.
"""
from collections import OrderedDict
from numbers import Number

import numpy as np
import pandas as pd

from omc3.utils import logging_tools, outliers
from omc3.definitions.constants import PLANES, PI2
from omc3.harpy.constants import (COL_TUNE, COL_AMP, COL_MU,
                                  COL_NATTUNE, COL_NATAMP, COL_NATMU,
                                  COL_FREQ, COL_PHASE)
from optics_functions.rdt import get_all_to_order

LOGGER = logging_tools.getLogger(__name__)

def _get_resonance_lines(order):
    resonances = {'X': [], 
                  'Y': [], 
                  'Z': [(1, 0, 1), (0, 1, 1), (1, 0, -1), (0, 1, -1)]}
    # Get all the rdts up to a given order
    fterms = get_all_to_order(order)
    # Some rdts can't be seen depending on the plane, filter them
    for (j,k,l,m) in fterms:
        if j != 0:
            resonances['X'].append((1-j+k, m-l, 0))
        if l != 0:
            resonances['Y'].append((k-j, 1-l+m, 0))
    return resonances


MAIN_LINES = {"X": (1, 0, 0), "Y": (0, 1, 0), "Z": (0, 0, 1)}
Z_TOLERANCE = 0.0003


def estimate_tunes(harpy_input, usvs):
    """
    Estimates the tunes from the `FFT` of decomposed data.

    Args:
        harpy_input: Analysis settings.
        usvs: dictionary per plane of (U and SV^T matrices).

    Returns:
        list of estimated tunes [x, y, z].
    """
    tunex = _estimate_tune(usvs["X"][1], harpy_input.window, q_s=False)
    tuney = _estimate_tune(usvs["Y"][1], harpy_input.window, q_s=False)
    if harpy_input.autotunes == "transverse":
        return [tunex, tuney, 0]
    tunez = _estimate_tune(usvs["X"][1], harpy_input.window, q_s=True)
    return [tunex, tuney, tunez]


def _estimate_tune(sv_mat, window, q_s=False):
    upper_power = int(np.power(2, np.ceil(np.log2(sv_mat.shape[1]))))
    synchro_limit = int(upper_power / 16)  # 0.03125: synchrotron tunes are lower, betatron higher
    coefs = np.abs(np.fft.rfft(np.mean(sv_mat, axis=0) * windowing(sv_mat.shape[1], window),
                               n=2 * upper_power))
    if q_s:
        return np.fft.rfftfreq(2 * upper_power)[np.argmax(coefs[:synchro_limit])]
    return np.fft.rfftfreq(2 * upper_power)[synchro_limit + np.argmax(coefs[synchro_limit:])]


def harpy_per_plane(harpy_input, bpm_matrix, usv, tunes, plane):
    """
    Calculates the spectra of TbT data, finds the main lines and cleans the BPMs for which the
    main line hasn't been found or is too far from its average over BPMs.

    Args:
        harpy_input: Analysis settings.
        bpm_matrix: TbT BPM matrix.
        usv: U and SV^T matrices decomposed matrices, can be ``None``.
        tunes: list of tunes [x, y, z].
        plane: marking the horizontal or vertical plane, **X** or **Y**.

    Returns:
        A tuple of DataFrame, Spectra, and Bad BPMs summary.
    """
    panda = pd.DataFrame(index=bpm_matrix.index, columns=OrderedDict())
    frequencies, coefficients = windowed_padded_rfft(harpy_input, bpm_matrix, tunes, usv)
    panda, not_tune_bpms = _get_main_resonances(tunes, dict(FREQS=frequencies, COEFFS=coefficients),
                                                plane, harpy_input.tolerance, panda)
    cleaned_by_tune_bpms = clean_by_tune(panda.loc[:, f"{COL_TUNE}{plane}"], harpy_input.tune_clean_limit)
    panda = panda.loc[panda.index.difference(cleaned_by_tune_bpms)]

    panda[f"{COL_MU}{plane}"] = _realign_phases(panda.loc[:, f"{COL_MU}{plane}"].to_numpy(),
                                          panda.loc[:, f"{COL_TUNE}{plane}"].to_numpy(), bpm_matrix.shape[1])

    bad_bpms_summaries = _get_bad_bpms_summary(not_tune_bpms, cleaned_by_tune_bpms)
    bpm_matrix = bpm_matrix.loc[panda.index]
    spectra = dict(FREQS=frequencies.loc[panda.index], COEFFS=coefficients.loc[panda.index])

    if _get_natural_tunes(harpy_input, tunes) is not None:
        panda = panda.join(_calculate_natural_tunes(spectra, _get_natural_tunes(harpy_input, tunes),
                                                    harpy_input.tolerance, plane))

        panda[f"{COL_NATMU}{plane}"] = _realign_phases(panda.loc[:, f"{COL_NATMU}{plane}"].to_numpy(),
                                                 panda.loc[:, f"{COL_NATTUNE}{plane}"].to_numpy(),

                                                 bpm_matrix.shape[1])
    if tunes[2] > 0:
        panda, _ = _get_main_resonances(tunes, spectra, "Z", Z_TOLERANCE, panda)
        panda[f"MUZ"] = _realign_phases(panda.loc[:, f"MUZ"].to_numpy(),
                                              panda.loc[:, f"TUNEZ"].to_numpy(),
                                              bpm_matrix.shape[1])
    return panda, spectra, bad_bpms_summaries


def find_resonances(tunes, nturns, plane, spectra, order_resonances):
    """
    Finds higher order lines in the spectra.

    Args:
        tunes: list of tunes [x, y, z].
        nturns: length of analysed data.
        plane: marking the horizontal or vertical plane, **X** or **Y**.
        spectra: frequencies and complex coefficients.

    Returns:
        A DataFrame.
    """
    resonance_lines = _get_resonance_lines(order_resonances)

    df = pd.DataFrame(index=spectra["FREQS"].index, columns=OrderedDict())
    resonances_freqs = _compute_resonances_with_freqs(plane, tunes, resonance_lines)
    if tunes[2] > 0.0:
        resonances_freqs.update(_compute_resonances_with_freqs("Z", tunes, resonance_lines))
    for resonance in resonances_freqs.keys():
        tolerance = _get_resonance_tolerance(resonance, nturns)
        max_coefs, max_freqs = _search_highest_coefs(resonances_freqs[resonance], tolerance,
                                                     spectra["FREQS"], spectra["COEFFS"])
        resstr = _get_resonance_suffix(resonance)
        df[f"{COL_FREQ}{resstr}"], df[f"{COL_AMP}{resstr}"], df[f"{COL_PHASE}{resstr}"] = _get_freqs_amps_phases(
            max_freqs, max_coefs, resonances_freqs[resonance])

        df[f"{COL_PHASE}{resstr}"] = _realign_phases(df.loc[:, f"{COL_PHASE}{resstr}"].to_numpy(),
                                               df.loc[:, f"{COL_FREQ}{resstr}"].to_numpy(), nturns)

    return df


def _get_main_resonances(tunes, spectra, plane, tolerance, df):
    freq = sum(r * t for r, t in zip(tunes, MAIN_LINES[plane])) % 1
    max_coefs, max_freqs = _search_highest_coefs(freq, tolerance, spectra["FREQS"], spectra["COEFFS"])
    if not np.any(max_coefs) and plane != "Z":
        raise ValueError(f"No main {plane} resonances found, "
                         f"try to increase the tolerance or adjust the tunes")
    bad_bpms_by_tune = spectra["COEFFS"].loc[max_coefs == 0.].index
    df[f"{COL_TUNE}{plane}"], df[f"{COL_AMP}{plane}"], df[f"{COL_MU}{plane}"] = _get_freqs_amps_phases(
        max_freqs, max_coefs, freq)
    if plane != "Z":
        df = df.loc[df.index.difference(bad_bpms_by_tune)]
    return df, bad_bpms_by_tune


def _calculate_natural_tunes(spectra, nattunes, tolerance, plane):
    df = pd.DataFrame(index=spectra["FREQS"].index, columns=OrderedDict())
    x, y, _ = nattunes
    freq = x % 1 if plane == "X" else y % 1
    max_coefs, max_freqs = _search_highest_coefs(freq, tolerance, spectra["FREQS"], spectra["COEFFS"])
    df[f"{COL_NATTUNE}{plane}"], df[f"{COL_NATAMP}{plane}"], df[f"{COL_NATMU}{plane}"] = _get_freqs_amps_phases(
        max_freqs, max_coefs, freq)
    return df


def _get_freqs_amps_phases(max_freqs, max_coefs, freq):
    return max_freqs, np.abs(max_coefs), np.sign(0.5 - freq) * np.angle(max_coefs) / PI2


def _realign_phases(phase_data, freq_data, nturns):
    mid_phase = (phase_data + freq_data * nturns / 2) % 1
    return np.where(np.abs(mid_phase) > 0.5, mid_phase - np.sign(mid_phase), mid_phase)


def clean_by_tune(tunes: pd.Series, tune_clean_limit: Number) -> pd.Series:
    """
    Looks for outliers in the tunes pandas Series and returns their indices.

    Args:
        tunes (pd.Series): Pandas series with the tunes per BPM and the BPM names as index.
        tune_clean_limit: No BPM will find as outlier if its distance to the
            average is lower than this limit.
    """
    bad_bpms_mask = outliers.get_filter_mask(tunes, limit=tune_clean_limit)  # returns ``True`` for good BPMs
    bad_bpms_names = tunes[~bad_bpms_mask].index
    return bad_bpms_names


def _get_bad_bpms_summary(not_tune_bpms, cleaned_by_tune_bpms):
    return ([f"{bpm_name} The main resonance has not been found" for bpm_name in not_tune_bpms] +
            [f"{bpm_name} tune is too far from average" for bpm_name in cleaned_by_tune_bpms])


def _search_highest_coefs(freq, tolerance, frequencies, coefficients):
    """
    Finds the highest coefficients in frequencies/coefficients in freq +- tolerance.

    Args:
        freq: frequency from interval (0, 1).
        tolerance:
        frequencies:
        coefficients:

    Returns:
        Tuple of maximum coefficients and the corresponding frequencies.
    """
    p_freq = freq if freq < 0.5 else 1 - freq
    min_val, max_val = p_freq - tolerance, p_freq + tolerance
    freq_vals = frequencies.to_numpy()
    coefs_vals = coefficients.to_numpy()
    on_window_mask = (freq_vals >= min_val) & (freq_vals <= max_val)
    filtered_coefs = np.where(on_window_mask, coefs_vals, 0)
    filtered_amps = np.abs(filtered_coefs)
    max_indices = np.argmax(filtered_amps, axis=1)
    max_coefs = filtered_coefs[np.arange(coefs_vals.shape[0]), max_indices]
    max_coefs = pd.Series(index=coefficients.index, data=max_coefs)
    max_pfreqs = freq_vals[np.arange(freq_vals.shape[0]), max_indices]
    max_freqs = max_pfreqs if freq < 0.5 else 1 - max_pfreqs
    max_freqs = pd.Series(index=coefficients.index, data=np.where(max_coefs != 0, max_freqs, 0))
    return max_coefs, max_freqs


def _get_resonance_suffix(resonance):
    x, y, z = resonance
    return f"{x}{y}{z if z else ''}".replace("-", "_")


def _compute_resonances_with_freqs(plane, tunes, resonance_lines):
    """
    Computes the frequencies in [0, 1) for all the resonances listed in the ``RESONANCE_LISTS``,
    together with the natural tunes frequencies if given.
    """
    freqs = [sum(r * t for r, t in zip(tunes, resonance)) % 1 for resonance in resonance_lines[plane]]
    return dict(zip(resonance_lines[plane], freqs))


def _compute_resonance_freqs(plane, tunes, resonance_lines):
    return [sum(r * t for r, t in zip(tunes, resonance)) % 1 for resonance in resonance_lines[plane]]


def _get_resonance_tolerance(resonance, n_turns):
    x, y, z = resonance
    return (abs(x) + abs(y) + bool(z) * (abs(z) - 1)) * max(1e-4, 1 / n_turns)


def windowed_padded_rfft(harpy_input, matrix, tunes, svd=None):
    """
    Calculates the spectra using specified windowing function and zero-padding.

    Args:
        harpy_input: A `HarpyInput` object.
        matrix: `pd.DataFrame` of TbT matrix (BPMs x turns).
        tunes: list of tunes [x, y, z].
        svd: reduced (U_matrix, np.dot(S_matrix, V_matrix)) of original TbT matrix, defaults to
            ``None``.

    Returns:
        Tuple of `pd.DataFrames`, for frequencies and coefficients.
    """
    padded_len, output_len = np.power(2, harpy_input.turn_bits), np.power(2, harpy_input.output_bits)
    sub_bins = int(padded_len / output_len)
    mask = get_freq_mask(harpy_input, tunes, 2 / matrix.shape[1])
    n_bins = int(np.sum(mask) / sub_bins)
    n_bpms = len(matrix.index)
    if svd is None:
        tbt_matrix = matrix.loc[:, :].to_numpy()
        coefs = np.fft.rfft(tbt_matrix * windowing(tbt_matrix.shape[1], window=harpy_input.window),
                            n=padded_len * 2)[:, mask]
    else:
        u, sv = svd
        s_vt_freq = np.fft.rfft(sv * windowing(sv.shape[1], window=harpy_input.window),
                                n=padded_len * 2)
        coefs = np.dot(u, s_vt_freq[:, mask])
    argsmax = (np.indices((n_bpms, n_bins))[1] * sub_bins +
               np.argmax(np.abs(np.reshape(coefs, (n_bpms, n_bins, sub_bins))), axis=2))
    # two 2 in following line is because we have just half of spectra
    coefficients = pd.DataFrame(index=matrix.index, data=2 * coefs[np.arange(n_bpms)[:, None], argsmax])
    del coefs
    frequencies = pd.DataFrame(index=matrix.index, data=np.outer(np.ones(n_bpms), np.fft.rfftfreq(
            padded_len * 2)[mask])[np.arange(n_bpms)[:, None], argsmax])
    return frequencies, coefficients


def windowing(length, window='hamming'):
    """
    Provides specified windowing function of given length.

    Currently, the following windowing functions are implemented (sorted by increasing width of
    main lobe, also decreasing spectral leakage in closest lobes):
    ``rectangle``, ``welch``, ``triangle``, ``hann``, ``hamming``, ``nuttal3``, and ``nuttal4``.

    Args:
        length: length of the window.
        window: type of the windowing function.

    Returns:
        Normalised windowing function of specified type and length.
    """
    ints2pi = PI2 * np.arange(length) / (length - 1)
    windows = {
        "nuttal4": 0.3125 - 0.46875 * np.cos(ints2pi) + 0.1875 * np.cos(2 * ints2pi) - 0.03125 * np.cos(3 * ints2pi),
        "nuttal3": 0.375 - 0.5 * np.cos(ints2pi) + 0.125 * np.cos(2 * ints2pi),
        "hamming": (25 / 46) - (21 / 46) * np.cos(ints2pi),
        "hann": 0.5 - 0.5 * np.cos(ints2pi),
        "welch": 1 - np.square((ints2pi / np.pi) - 1),
        "triangle": 1 - np.abs((ints2pi / np.pi) - 1),
        "rectangle": np.ones(length)
    }
    if window not in windows.keys():
        raise NotImplementedError(f"Unknown windowing function {window}")
    return windows[window] / np.sum(windows[window])


def get_freq_mask(harpy_input, tunes, auto_tol):
    """
    Computes mask to get intervals around resonances in frequency domain.

    Args:
        harpy_input: A `HarpyInput` object.
        tunes: list of tunes [x, y, z].
        auto_tol: automatically calculated tolerance.

    Returns:
        Boolean array.
    """
    if "full_spectra" in harpy_input.to_write:
        mask = np.ones(int(np.power(2, harpy_input.turn_bits)) + 1, dtype=bool)
        mask[-1] = False
        return mask
    mask = np.zeros(int(np.power(2, harpy_input.turn_bits)) + 1, dtype=bool)
    nattunes = _get_natural_tunes(harpy_input, tunes)
    if nattunes is not None:
        mask = _get_partial_freq_mask(harpy_input, mask, list(nattunes), harpy_input.tolerance)
    tol = harpy_input.tolerance if harpy_input.autotunes is None else auto_tol
    freqs = (list(tunes))
    
    resonance_lines = _get_resonance_lines(harpy_input.resonances)
    for plane in PLANES:
        freqs.extend(_compute_resonance_freqs(plane, tunes, resonance_lines))
    if tunes[2]:
        freqs.extend(_compute_resonance_freqs("Z", tunes, resonance_lines))
    return _get_partial_freq_mask(harpy_input, mask, freqs, tol)


def _get_natural_tunes(harpy_input, tunes):
    if harpy_input.natdeltas is not None:
        return [r + t if t != 0 else 0 for r, t in zip(tunes, harpy_input.natdeltas)]
    if harpy_input.nattunes is not None:
        return harpy_input.nattunes
    return None


def _get_partial_freq_mask(harpy_input, start_mask, frequencies, tolerance):
    bins = np.power(2, harpy_input.output_bits)
    sub_bins = np.power(2, harpy_input.turn_bits - harpy_input.output_bits)
    mask = start_mask[:]
    freqs = np.array(frequencies) % 1
    freqs = np.where(freqs > 0.5, 1 - freqs, freqs)
    for freq in freqs:
        if freq == 0.0:
            continue
        mask[int(np.floor(max(0, freq - tolerance) * 2 * bins) * sub_bins):
             int(np.ceil(min(0.5, freq + tolerance) * 2 * bins) * sub_bins)] = True
    mask[-1] = False  # 2^n + 1 is undesired
    return mask



"""
Helper
------

This module contains helper functionality for ``kmod``.
It provides functions to perform data cleaning, IO loading and plotting.
"""
import datetime

import numpy as np
import tfs
from matplotlib import pyplot as plt

from omc3.definitions import formats
from omc3.definitions.constants import PLANES
from omc3.kmod import analysis
from omc3.kmod.constants import SIDES, ERR, TUNE, EXT, CLEANED, K, AVERAGE, BETA
from omc3.utils import logging_tools, outliers

LOG = logging_tools.get_logger(__name__)


def clean_data(magnet_df, no_autoclean):
    if no_autoclean:
        LOG.info('Manual cleaning is not yet implemented, no cleaning was performed')
        for plane in PLANES:
            magnet_df[f"{CLEANED}{plane}"] = True
    else:
        LOG.debug('Automatic Tune cleaning')
        for plane in PLANES:
            magnet_df[f"{CLEANED}{plane}"] = outliers.get_filter_mask(
                magnet_df[f"{TUNE}{plane}"].to_numpy(), x_data=magnet_df[K].to_numpy(), limit=1e-5)
    return magnet_df


def add_tune_uncertainty(magnet_df, tune_uncertainty):
    LOG.debug(f'adding {tune_uncertainty} units tune measurement uncertainty')
    for plane in PLANES:
        magnet_df[f"{ERR}{TUNE}{plane}"] = np.sqrt(magnet_df[f"{ERR}{TUNE}{plane}"]**2 + tune_uncertainty**2)
    return magnet_df


# ##########################   FILE LOADING    ##########################################


def get_input_data(opt):
    return get_simulation_files(opt.working_directory, opt.beam, opt.magnets) if opt.simulation else merge_data(opt)


def get_simulation_files(working_directory, beam, magnets):
    magnet1_df = tfs.read(working_directory / f"{magnets[0]}.B{beam:d}{EXT}")
    magnet2_df = tfs.read(working_directory / f"{magnets[1]}.B{beam:d}{EXT}")
    magnet1_df.headers['QUADRUPOLE'] = magnet1_df.headers['NAME']
    magnet2_df.headers['QUADRUPOLE'] = magnet2_df.headers['NAME']
    return magnet1_df, magnet2_df


def merge_data(kmod_input_params):
    magnet_df = []
    work_dir = kmod_input_params.working_directory
    ip = kmod_input_params.interaction_point
    beam = kmod_input_params.beam
    for (filepaths, magnet) in zip(
            return_ip_filename(work_dir, ip, beam) if ip is not None
            else return_circuit_filename(work_dir, kmod_input_params.circuits, beam),
            kmod_input_params.magnets):
        LOG.debug(f'Loading tunes from {filepaths[0]} and {filepaths[1]}')
        tune_dfs = dict(X=tfs.read(filepaths[0]), Y=tfs.read(filepaths[1]))
        LOG.debug(f'Loading k from {filepaths[2]}')
        k_df = tfs.read(filepaths[2])
        LOG.debug('Binning data')
        magnet_df.append(bin_tunes_and_k(tune_dfs, k_df, magnet))
    return magnet_df


def return_ip_filename(working_directory, ip, beam):
    LOG.debug('Setting IP trim file names')
    all_filepaths = []
    for side in SIDES:
        path_tunex = working_directory / f"{ip.lower()}b{beam:d}{side}X{EXT}"
        path_tuney = working_directory / f"{ip.lower()}b{beam:d}{side}Y{EXT}"
        path_k = working_directory / f"{ip.lower()}{side}K{EXT}"
        all_filepaths.append([path_tunex, path_tuney, path_k])
    return all_filepaths


def return_circuit_filename(working_directory, circuits_1_and_2, beam):
    LOG.debug('Setting Circuit trim file names')
    all_filepaths = []
    for circuit in circuits_1_and_2:
        path_tunex = working_directory / f"{circuit}_tune_x_b{beam}{EXT}"
        path_tuney = working_directory / f"{circuit}_tune_y_b{beam}{EXT}"
        path_k = working_directory / f"{circuit}_k{EXT}"
        all_filepaths.append([path_tunex, path_tuney, path_k])
    return all_filepaths


def bin_tunes_and_k(tune_dfs, k_df, magnet):
    # create bins, centered around each time step in k with width eq half distance to the next timestep
    bins = np.append((k_df['TIME']-k_df.diff()['TIME']/2.).fillna(value=0).values, k_df['TIME'].iloc[-1])
    magnet_df = k_df.loc[:, ['K']]
    magnet_df['K'] = np.abs(magnet_df['K'].to_numpy())
    for plane in PLANES:
        magnet_df[f"{TUNE}{plane}"], magnet_df[f"{ERR}{TUNE}{plane}"] = return_mean_of_binned_data(bins, tune_dfs[plane])
    return tfs.TfsDataFrame(magnet_df, headers=headers_for_df(magnet, k_df))


def return_mean_of_binned_data(bins, tune_df):
    digitize = np.digitize(tune_df['TIME'], bins)
    mean = [tune_df['TUNE'][digitize == i].mean() for i in range(1, len(bins))]
    std = np.nan_to_num([tune_df['TUNE'][digitize == i].std() for i in range(1, len(bins))])
    return mean, std


def headers_for_df(magnet, k_df):
    LOG.debug('Creating headers for DF')
    head = {}
    head['QUADRUPOLE'] = magnet
    head['DELTA_I'] = np.max(k_df['CURRENT'].rolling(5).mean()) - np.min(k_df['CURRENT'].rolling(5).mean()) / 2
    head['START_TIME'] = datetime.datetime.fromtimestamp(k_df['TIME'].iloc[0] / 1000).strftime(formats.TIME)
    head['END_TIME'] = datetime.datetime.fromtimestamp(k_df['TIME'].iloc[-1] / 1000).strftime(formats.TIME)
    # add starting tunes/tunesplit, number of cycles, ... to header
    return head


# ##############################    PLOTING    #############################################


def plot_cleaned_data(magnet_dfs, plot_name, interactive_plot=False):
    fig, ax = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))
    for i, plane in enumerate(PLANES):
        for j in range(2):
            ax_plot(ax[i, j], magnet_dfs[j], plane)
    ax[1, 1].legend()
    plt.tight_layout()
    plt.savefig(plot_name)
    if interactive_plot:
        plt.show()
    return


def ax_plot(ax, magnet_df, plane):
    ax.set_title(magnet_df.headers['QUADRUPOLE'], fontsize=15)
    ax_errorbar_plot(ax=ax, magnet_df=magnet_df, plane=plane, clean=True,
                     plot_settings={"color": "blue", "marker": "o", "label": "Data", "zorder": 1})
    ax_errorbar_plot(ax=ax, magnet_df=magnet_df, plane=plane, clean=False,
                     plot_settings={"color": "orange", "marker": "o", "label": "Cleaned", "zorder": 2})
    ax.plot((magnet_df.where(magnet_df[f"{CLEANED}{plane}"])[K].dropna() - magnet_df.headers[K]) * 1E3,
            analysis.fit_prec(
                analysis.return_fit_input(magnet_df, plane),
                magnet_df.headers[f"{AVERAGE}{BETA}{plane}"]) + magnet_df.headers[f"{TUNE}{plane}"],
            color='red', label='Fit', zorder=3)
    ax.set_xlabel(r'$ \Delta K $', fontsize=15)
    ax.set_ylabel(r'$ Q_{{{:s}}} $'.format(plane), fontsize=15)
    return


def ax_errorbar_plot(ax, magnet_df, plane, clean, plot_settings):
    new_df = magnet_df.loc[magnet_df.loc[:, f"{CLEANED}{plane}"].to_numpy() == clean, :]
    ax.errorbar((new_df.loc[:, K].dropna() - magnet_df.headers[K]) * 1E3,
                new_df.loc[:, f"{TUNE}{plane}"].dropna(),
                yerr=new_df.loc[:, f"{ERR}{TUNE}{plane}"].dropna(),
                color=plot_settings["color"],
                fmt=plot_settings["marker"],
                label=plot_settings["label"],
                zorder=plot_settings["zorder"]
                )
    return ax



"""
Hole in One
-----------

``hole_in_one`` is the top-level script of analysis functionality offered in ``omc3``. In most of
your use cases, this is the file you will want to call. It handles:
- frequency spectra of Turn-by-Turn BPM data,
- various lattice optics parameters from frequency spectra,
- various lattice optics parameters from Turn-by-Turn BPM data,

A general analysis workflow, from straight out turn-by-turn measurement or simulations files to
results, goes as follows:

+-----------------------+--------+---------------------+------+-----------------------------------+
|                      Analysis Workflow                                                          |
+=======================+========+=====================+======+===================================+
| Turn-by-Turn BPM data | --->   |  frequency spectra  | ---> | various lattice optics parameters |
+-----------------------+--------+---------------------+------+-----------------------------------+

The first step above consists in frequency analysis performed by ``harpy``, while the second
one is optics analysis performed by ``measure_optics``. Each corresponding stage is represented
by a different set of files:

+--------------------------+--------+---------------------------+------+-----------------------+
|                     Corresponding Files                                                      |
+==========================+========+===========================+======+=======================+
|  SDDS file:  **.sdds**   | --->   |  Tfs files: **.lin[xy]**  | ---> |  Tfs files: **.tfs**  |
+--------------------------+--------+---------------------------+------+-----------------------+

To run either of the two or both steps, see options ``--harpy`` and ``--optics``.
"""
import os
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from os.path import abspath, basename, dirname, join

from generic_parser.entrypoint_parser import (EntryPoint, EntryPointParameters,
                                              add_to_arguments, entrypoint,
                                              save_options_to_config)

import turn_by_turn as tbt
from omc3.definitions import formats
from omc3.harpy import handler
from omc3.model import manager
from omc3.optics_measurements import measure_optics
from omc3.utils import iotools, logging_tools
from omc3.utils.contexts import timeit

LOGGER = logging_tools.get_logger(__name__)

DEFAULT_CONFIG_FILENAME = "analysis_{time:s}.ini"


def hole_in_one_params():
    params = EntryPointParameters()
    params.add_parameter(name="harpy", action="store_true", help="Runs frequency analysis")
    params.add_parameter(name="optics", action="store_true", help="Measures the lattice optics")
    return params


@entrypoint(hole_in_one_params(), strict=False)
def hole_in_one_entrypoint(opt, rest):
    """
    Runs frequency analysis and measures lattice optics.


    Hole_in_one Kwargs:
      - **harpy**: Runs frequency analysis

        Flags: **--harpy**
        Action: ``store_true``

      - **optics**: Measures the lattice optics
        Flags: **--optics**
        Action: ``store_true``

    Harpy Kwargs:
      - **files**: TbT files to analyse

        Flags: **--files**
        Required: ``True``
      - **outputdir**: Output directory.

        Flags: **--outputdir**
        Required: ``True``
      - **to_write**: Choose the type of output.

        Flags: **--to_write**
        Choices: ``('lin', 'spectra', 'full_spectra', 'bpm_summary')``
        Default: ``['lin', 'bpm_summary']``
      - **turns** *(int)*: Turn index to start and first turn index to be ignored.

        Flags: **--turns**
        Default: ``[0, 50000]``
      - **unit** *(str)*: A unit of TbT BPM orbit data. All cuts and output are in 'm'.

        Flags: **--unit**
        Choices: ``('m', 'cm', 'mm', 'um')``
        Default: ``m``
      - **tbt_datatype** *(str)*: Choose datatype from which to import (e.g LHC binary SDDS, numpy npz).

        Flags: **--tbt_datatype**
        Default: ``LHC``

      *--Cleaning--*

      - **clean**: If present, the data are first cleaned.

        Flags: **--clean**
        Action: ``store_true``
      - **bad_bpms**: Bad BPMs to clean.

        Flags: **--bad_bpms**
      - **first_bpm** *(str)*: First BPM in the measurement.
        Used to resynchronise the TbT data with model.

        Flags: **--first_bpm**
      - **keep_exact_zeros**: If present, will not remove BPMs with exact zeros in TbT data.

        Flags: **--keep_exact_zeros**
        Action: ``store_true``
      - **max_peak** *(float)*: Removes BPMs where the maximum orbit > limit.

        Flags: **--max_peak**
        Default: ``0.02``
      - **model**: Model for BPM locations

        Flags: **--model**
      - **num_svd_iterations** *(int)*: Maximal number of iterations of U matrix elements removal
        and renormalisation in iterative SVD cleaning of dominant BPMs.
        This is also equal to maximal number of BPMs removed per SVD mode.

        Flags: **--num_svd_iterations**
        Default: ``3``
      - **opposite_direction**: If present, beam in the opposite direction to model
        is assumed for resynchronisation of BPMs.

        Flags: **--opposite_direction**
        Action: ``store_true``
      - **peak_to_peak** *(float)*: Peak to peak amplitude cut. This removes BPMs,
        where abs(max(turn values) - min(turn values)) <= threshold.

        Flags: **--peak_to_peak**
        Default: ``1e-08``
      - **sing_val** *(int)*: Keep this amount of largest singular values.

        Flags: **--sing_val**
        Default: ``12``
      - **svd_dominance_limit** *(float)*: Limit for single BPM dominating a mode.

        Flags: **--svd_dominance_limit**
        Default: ``0.925``
      - **wrong_polarity_bpms**: BPMs with swapped polarity in both planes.

        Flags: **--wrong_polarity_bpms**

      *--Frequency Analysis--*

      - **autotunes** *(str)*: The main tunes are guessed as the strongest line in SV^T matrix
        frequency spectrum: Synchrotron tune below ~0.03, betatron tunes above ~0.03.

        Flags: **--autotunes**
        Choices: ``('all', 'transverse')``
      - **is_free_kick**: If present, it will perform the free kick phase correction

        Flags: **--free_kick**
        Action: ``store_true``
      - **natdeltas** *(float)*: Guess for the offsets of natural tunes from
        the driven tunes (x, y, z). Disabled when set to 0.

        Flags: **--natdeltas**
      - **nattunes** *(float)*: Guess for the natural tunes (x, y, z).  Disabled when set to 0.

        Flags: **--nattunes**
      - **output_bits** *(int)*: Number (frequency, complex coefficient) pairs in the output
        is up to 2 ** output_bits (maximal in case full spectra is output).
        There is one pair (with maximal amplitude of complex coefficient) per interval
        of size 2 ** (- output_bits - 1).

        Flags: **--output_bits**
        Default: ``12``
      - **tolerance** *(float)*: Tolerance specifying an interval in frequency domain,
        where to look for the tunes.

        Flags: **--tolerance**
        Default: ``0.01``
      - **tune_clean_limit** *(float)*: The tune cleaning wont remove BPMs because of measured
        tune outliers closer to the average tune than this limit.

        Flags: **--tune_clean_limit**
        Default: ``1e-05``
      - **tunes** *(float)*: Guess for the main tunes [x, y, z]. Tunez is disabled when set to 0

        Flags: **--tunes**
      - **turn_bits** *(int)*: Number (frequency, complex coefficient) pairs in the calculation
        is 2 ** turn_bits, i.e. the difference between two neighbouring frequencies
        is 2 ** (- turn_bits - 1).

        Flags: **--turn_bits**
        Default: ``20``
      - **window** *(str)*: Windowing function to be used for frequency analysis.

        Flags: **--window**
        Choices: ``('rectangle', 'welch', 'triangle', 'hann', 'hamming', 'nuttal3', 'nuttal4')``
        Default: ``hann``
      - **resonances** *(int)*: Maximum magnet order of resonance lines to calculate.

        Flags: **--resonances**
        Choices: ``(2 <= n <= 8)``
        Default: ``4``


    Optics Kwargs:
      - **files**: Files for analysis

        Flags: **--files**
        Required: ``True``
      - **outputdir**: Output directory

        Flags: **--outputdir**
        Required: ``True``
      - **calibrationdir** *(str)*: Path to calibration files directory.

        Flags: **--calibrationdir**
      - **chromatic_beating**: Calculate chromatic beatings: W, PHI and coupling

        Flags: **--chromatic_beating**
        Action: ``store_true``
      -  **compensation** *(str)*: Mode of compensation for the analysis after driven beam excitation.

        Flags: **-compensation**
        Choices: ``("model", "equation", "none")``
        Default: ``model``
      - **coupling_method** *(int)*: Coupling analysis option: disabled, 1 BPM or 2 BPMs method

        Flags: **--coupling_method**
        Choices: ``(0, 1, 2)``
        Default: ``2``
      - **coupling_pairing**: Pairing mode for 2 BPM coupling method. If 0 is given, omc3 
        will try to determine the best candidate. If a number n>=1 is given, then some BPMs are 
        skipped and the n-th following BPM downstream is used for the pairing.

        Flags: **--coupling_pairing**
        Choices: ``(0, n>=1)``
        Default: ``0``.
      - **nonlinear**: Calculate higher order RDTs or CRDT

        Flags: **--nonlinear**
        Choices: ``(rdt, crdt)``
        Default: ``None``
      - **rdt_magnet_order**: Maximum magnet order for RDTs calculation if --nonlinear is given

        Flags: **--rdt_magnet_order**
        Choices: ``(2 <= n <= 8)``
        Default: ``4``
      - **only_coupling**: Calculate only coupling.

        Flags: **--only_coupling**
        Action: ``store_true``
      - **range_of_bpms** *(int)*: Range of BPMs for beta from phase calculation

        Flags: **--range_of_bpms**
        Choices: ``(5, 7, 9, 11, 13, 15)``
        Default: ``11``
      - **three_bpm_method**: Use 3 BPM method in beta from phase

        Flags: **--three_bpm_method**
        Action: ``store_true``

      - **three_d_excitation**: Use 3D kicks to calculate dispersion.
        Action: ``store_true``

      - **isolation_forest**: If present, remove outlying BPMs with isolation forest.

        Flags: **--isolation_forest**
        Action: ``store_true``

      - **second_order_dispersion**: If present, calculate second order dispersion.

        Flags: **--second_order_dispersion**
        Action: ``store_true``

      - **union**: If present, the phase advances are calculate for union of BPMs
        with at least 3 valid measurements, instead of intersection .

        Flags: **--union**
        Action: ``store_true``


    Accelerator Kwargs:
      - **accel**: Choose the accelerator to use. More details can be found in omc3/model/manager.py

        Flags: **--accel**
        Required: ``True``
      - **model_dir**: Model directory, specify if ``--model`` option is not used.

        Flags: **--model_dir**

      - For the rest, please see get_parameters() methods in child Accelerator classes,
        which are declared in ``omc3/model/accelerators/*.py``.
    """
    if not opt.harpy and not opt.optics:
        raise SystemError("No module has been chosen.")
    if not rest:
        raise SystemError("No input has been set.")
    harpy_opt, optics_opt, accel_opt = _get_suboptions(opt, rest)
    _write_config_file(harpy_opt, optics_opt, accel_opt)
    lins = []
    if harpy_opt is not None:
        lins = _run_harpy(harpy_opt)
    if optics_opt is not None:
        _measure_optics(lins, optics_opt)


def _get_suboptions(opt, rest):
    if opt.harpy:
        harpy_opt, rest = _harpy_entrypoint(rest)
        if opt.optics:
            rest = add_to_arguments(rest, entry_params=optics_params(),
                                    files=harpy_opt.files,
                                    outputdir=harpy_opt.outputdir)
            harpy_opt.outputdir = join(harpy_opt.outputdir, 'lin_files')
            if harpy_opt.model is not None:
                rest = add_to_arguments(rest, entry_params={"model_dir": {"flags": "--model_dir"}},
                                        model_dir=dirname(abspath(harpy_opt.model)))
    else:
        harpy_opt = None

    if opt.optics:
        optics_opt, rest = _optics_entrypoint(rest)
        accel_opt = manager.get_parsed_opt(rest)
        optics_opt.accelerator = manager.get_accelerator(rest)
        if not optics_opt.accelerator.excitation and optics_opt.compensation != "none":
            raise AttributeError("Compensation requested and no driven model was provided.")
    else:
        optics_opt = None
        accel_opt = None
    return harpy_opt, optics_opt, accel_opt


def _write_config_file(harpy_opt, optics_opt, accelerator_opt):
    """Write the parsed options into a config file for later use."""
    all_opt = OrderedDict()
    if harpy_opt is not None:
        all_opt["harpy"] = True
        all_opt.update(OrderedDict(sorted(harpy_opt.items())))

    if optics_opt is not None:
        optics_opt = OrderedDict(sorted(optics_opt.items()))
        optics_opt.pop('accelerator')

        all_opt["optics"] = True
        all_opt.update(optics_opt)
        all_opt.update(sorted(accelerator_opt.items()))

    out_dir = all_opt["outputdir"]
    file_name = DEFAULT_CONFIG_FILENAME.format(time=datetime.utcnow().strftime(formats.TIME))
    iotools.create_dirs(out_dir)

    save_options_to_config(os.path.join(out_dir, file_name), all_opt)


def _run_harpy(harpy_options):
    iotools.create_dirs(harpy_options.outputdir)
    with timeit(lambda spanned: LOGGER.info(f"Total time for Harpy: {spanned}")):
        lins = []
        all_options = _replicate_harpy_options_per_file(harpy_options)
        tbt_datas = [(tbt.read_tbt(option.files, datatype=option.tbt_datatype), option) for option in all_options]
        for tbt_data, option in tbt_datas:
            lins.extend([handler.run_per_bunch(bunch_data, bunch_options)
                         for bunch_data, bunch_options in _multibunch(tbt_data, option)])
    return lins


def _replicate_harpy_options_per_file(options):
    list_of_options = []
    for input_file in options.files:
        new_options = deepcopy(options)
        new_options.files = input_file
        list_of_options.append(new_options)
    return list_of_options


def _multibunch(tbt_datas, options):
    if tbt_datas.nbunches == 1:
        yield tbt_datas, options
        return
    for index in range(tbt_datas.nbunches):
        new_options = deepcopy(options)
        new_file_name = f"bunchid{tbt_datas.bunch_ids[index]}_{basename(new_options.files)}"
        new_options.files = join(dirname(options.files), new_file_name)
        yield tbt.TbtData([tbt_datas.matrices[index]], tbt_datas.date,
                          [tbt_datas.bunch_ids[index]], tbt_datas.nturns), new_options


def _measure_optics(lins, optics_opt):
    if len(lins) == 0:
        lins = optics_opt.files
    inputs = measure_optics.InputFiles(lins, optics_opt)
    iotools.create_dirs(optics_opt.outputdir)
    calibrations = measure_optics.copy_calibration_files(optics_opt.outputdir,
                                                         optics_opt.calibrationdir)
    inputs.calibrate(calibrations)
    with timeit(lambda spanned: LOGGER.info(f"Total time for optics measurements: {spanned}")):
        measure_optics.measure_optics(inputs, optics_opt)


def _harpy_entrypoint(params):
    options, rest = EntryPoint(harpy_params(), strict=False).parse(params)
    if options.natdeltas is not None and options.nattunes is not None:
        raise AttributeError("Colliding options found: --nattunes and --natdeltas. Choose only one")
    if options.tunes is not None and options.autotunes is not None:
        raise AttributeError("Colliding options found: --tunes and --autotunes. Choose only one")
    if options.tunes is None and options.autotunes is None:
        raise AttributeError("One of the options --tunes and --autotunes has to be used.")
    if options.svd_dominance_limit <= 0.0:
        raise AttributeError("SVD dominance limit should be positive")
    if options.bad_bpms is None:
        options.bad_bpms = []
    if options.wrong_polarity_bpms is None:
        options.wrong_polarity_bpms = []
    if options.is_free_kick:
        options.window = "rectangle"
    if not 2 <= options.resonances <= 8:
        raise AttributeError("The magnet order for resonance lines calculation should be between 2 and 8 (inclusive).")

    return options, rest


def harpy_params():
    params = EntryPointParameters()
    params.add_parameter(name="files", required=True, nargs='+', help="TbT files to analyse")
    params.add_parameter(name="outputdir", required=True, help="Output directory.")
    params.add_parameter(name="model", help="Model for BPM locations")
    params.add_parameter(name="unit", type=str, default=HARPY_DEFAULTS["unit"],
                         choices=("m", "cm", "mm", "um"),
                         help=f"A unit of TbT BPM orbit data. All cuts and output are in 'm'.")
    params.add_parameter(name="turns", type=int, nargs=2, default=HARPY_DEFAULTS["turns"],
                         help="Turn index to start and first turn index to be ignored.")
    params.add_parameter(name="to_write", nargs='+', default=HARPY_DEFAULTS["to_write"],
                         choices=('lin', 'spectra', 'full_spectra', 'bpm_summary'),
                         help="Choose the type of output.")
    params.add_parameter(name="tbt_datatype", default=HARPY_DEFAULTS["tbt_datatype"],
                         choices=list(tbt.io.TBT_MODULES.keys()),
                         help="Choose the datatype from which to import. ")

    # Cleaning parameters
    params.add_parameter(name="clean", action="store_true",
                         help="If present, the data are first cleaned.")
    params.add_parameter(name="sing_val", type=int, default=HARPY_DEFAULTS["sing_val"],
                         help="Keep this amount of largest singular values.")
    params.add_parameter(name="peak_to_peak", type=float, default=HARPY_DEFAULTS["peak_to_peak"],
                         help="Peak to peak amplitude cut. This removes BPMs, "
                              "where abs(max(turn values) - min(turn values)) <= threshold.")
    params.add_parameter(name="max_peak", type=float, default=HARPY_DEFAULTS["max_peak"],
                         help="Removes BPMs where the maximum orbit > limit.")
    params.add_parameter(name="svd_dominance_limit", type=float,
                         default=HARPY_DEFAULTS["svd_dominance_limit"],
                         help="Limit for single BPM dominating a mode.")
    params.add_parameter(name="num_svd_iterations", type=int,
                         default=HARPY_DEFAULTS["num_svd_iterations"],
                         help="Maximal number of iterations of U matrix elements removal "
                              "and renormalisation in iterative SVD cleaning of dominant BPMs."
                              " This is also equal to maximal number of BPMs removed per SVD mode.")
    params.add_parameter(name="bad_bpms", nargs='*', help="Bad BPMs to clean.")
    params.add_parameter(name="wrong_polarity_bpms", nargs='*',
                         help="BPMs with swapped polarity in both planes.")
    params.add_parameter(name="keep_exact_zeros", action="store_true",
                         help="If present, will not remove BPMs with exact zeros in TbT data.")
    params.add_parameter(name="first_bpm", type=str,
                         help="First BPM in the measurement. "
                              "Used to resynchronise the TbT data with model.")
    params.add_parameter(name="opposite_direction", action="store_true",
                         help="If present, beam in the opposite direction to model"
                              " is assumed for resynchronisation of BPMs.")

    # Harmonic analysis parameters
    params.add_parameter(name="tunes", type=float, nargs=3,
                         help="Guess for the main tunes [x, y, z]. Tunez is disabled when set to 0")
    params.add_parameter(name="nattunes", type=float, nargs=3,
                         help="Guess for the natural tunes (x, y, z).  Disabled when set to 0.")
    params.add_parameter(name="natdeltas", type=float, nargs=3,
                         help="Guess for the offsets of natural tunes from the driven tunes"
                              " (x, y, z). Disabled when set to 0.")
    params.add_parameter(name="autotunes", type=str, choices=("all", "transverse"),
                         help="The main tunes are guessed as "
                              "the strongest line in SV^T matrix frequency spectrum: "
                              "Synchrotron tune below ~0.03, betatron tunes above ~0.03.")
    params.add_parameter(name="tune_clean_limit", type=float,
                         default=HARPY_DEFAULTS["tune_clean_limit"],
                         help="The tune cleaning wont remove BPMs because of measured tune outliers"
                              " closer to the average tune than this limit.")
    params.add_parameter(name="tolerance", type=float,
                         default=HARPY_DEFAULTS["tolerance"],
                         help="Tolerance specifying an interval in frequency domain, where to look "
                              "for the tunes.")
    params.add_parameter(name="is_free_kick", action="store_true",
                         help="If present, it will perform the free kick phase correction")
    params.add_parameter(name="window", type=str, default=HARPY_DEFAULTS["window"],
                         choices=("rectangle", "hann", "triangle", "welch", "hamming", "nuttal3",
                                  "nuttal4"),
                         help="Windowing function to be used for frequency analysis.")
    params.add_parameter(name="turn_bits", type=int, default=HARPY_DEFAULTS["turn_bits"],
                         help="Number (frequency, complex coefficient) pairs in the calculation"
                              " is 2 ** turn_bits, i.e. the difference between "
                              "two neighbouring frequencies is 2 ** (- turn_bits - 1).")
    params.add_parameter(name="output_bits", type=int, default=HARPY_DEFAULTS["output_bits"],
                         help="Number (frequency, complex coefficient) pairs in the output "
                              "is up to 2 ** output_bits (maximal in case full spectra is output). "
                              "There is one pair (with maximal amplitude of complex coefficient) "
                              "per interval of size 2 ** (- output_bits - 1).")
    params.add_parameter(name="resonances", type=int, default=HARPY_DEFAULTS["resonances"],
                        help="Maximum magnet order of resonance lines to calculate.")
    return params


def _optics_entrypoint(params):
    options, rest = EntryPoint(optics_params(), strict=False).parse(params)
    
    if "rdt" in options.nonlinear and not 2 <= options.rdt_magnet_order <= 8:
        raise AttributeError("The magnet order for RDT calculation should be between 2 and 8 (inclusive).")

    return options, rest


def optics_params():
    params = EntryPointParameters()
    params.add_parameter(name="files", required=True, nargs='+',
                         help="Files for analysis")
    params.add_parameter(name="outputdir", required=True,
                         help="Output directory")
    params.add_parameter(name="calibrationdir", type=str,
                         help="Path to calibration files directory.")
    params.add_parameter(name="coupling_method", type=int,
                         choices=(0, 1, 2), default=OPTICS_DEFAULTS["coupling_method"],
                         help="Analysis option for coupling: disabled, 1 BPM or 2 BPMs method")
    params.add_parameter(name="coupling_pairing", type=int,
                         default=OPTICS_DEFAULTS["coupling_pairing"],
                         help="Pairing mode for 2 BPM coupling method. If 0 is given, omc3 will try to "
                              "determine the best candidate. If a number n>=1 is given, then some BPMs are skipped "
                              "and the n-th following BPM downstream is used for the pairing.")
    params.add_parameter(name="range_of_bpms", type=int,
                         choices=(5, 7, 9, 11, 13, 15),  default=OPTICS_DEFAULTS["range_of_bpms"],
                         help="Range of BPMs for beta from phase calculation")
    params.add_parameter(name="union", action="store_true",
                         help="If present, the phase advances are calculate for union of BPMs "
                              "with at least 3 valid measurements, instead of intersection .")
    params.add_parameter(name="nonlinear", nargs='*', default=[],
                         choices=('rdt', 'crdt'),
                         help="Choose which rdt analysis is conducted.")
    params.add_parameter(name="rdt_magnet_order", type=int, default=OPTICS_DEFAULTS["rdt_magnet_order"],
                         help="Maximum magnet order for the RDT calculation.")
    params.add_parameter(name="three_bpm_method", action="store_true",
                         help="Use 3 BPM method in beta from phase")
    params.add_parameter(name="only_coupling", action="store_true", help="Calculate only coupling. ")
    params.add_parameter(name="compensation", type=str, default=OPTICS_DEFAULTS["compensation"],
                         choices=("model", "equation", "none"),
                         help="Mode of compensation for the analysis after driven beam excitation")
    params.add_parameter(name="three_d_excitation", action="store_true",
                         help="Use 3D kicks to calculate dispersion")
    params.add_parameter(name="isolation_forest", action="store_true",
                         help="Remove outlying BPMs with isolation forest")
    params.add_parameter(name="second_order_dispersion", action="store_true",
                         help="Calculate second order dispersion")
    params.add_parameter(name="chromatic_beating", action="store_true",
                         help="Calculate chromatic beatings: W, PHI and coupling")
    return params


HARPY_DEFAULTS = {
    "turns": [0, 50000],
    "unit": "m",
    "sing_val": 12,
    "peak_to_peak": 1e-8,
    "max_peak": 0.02,
    "svd_dominance_limit": 0.925,
    "num_svd_iterations": 3,
    "tolerance": 0.01,
    "tune_clean_limit": 1e-5,
    "window": "hann",
    "turn_bits": 20,
    "output_bits": 12,
    "to_write": ["lin", "bpm_summary"],
    "tbt_datatype": "lhc",
    "resonances": 4,
}

OPTICS_DEFAULTS = {
        "coupling_method": 2,
        "coupling_pairing": 0,
        "range_of_bpms": 11,
        "compensation": "model",
        "rdt_magnet_order": 4,
}


if __name__ == "__main__":
    hole_in_one_entrypoint()



import pathlib

import setuptools

# The directory containing this file
TOPLEVEL_DIR = pathlib.Path(__file__).parent.absolute()
ABOUT_FILE = TOPLEVEL_DIR / "omc3" / "__init__.py"
README = TOPLEVEL_DIR / "README.md"

# Information on the omc3 package
ABOUT_OMC3: dict = {}
with ABOUT_FILE.open("r") as f:
    exec(f.read(), ABOUT_OMC3)

with README.open("r") as docs:
    long_description = docs.read()

# Dependencies for the package itself
DEPENDENCIES = [
    "matplotlib>=3.2.0,<3.4.3",  # temporary limitations, try run test!
    "Pillow>=6.2.2",  # not our dependency but older versions crash with mpl
    "numpy>=1.19.0",
    "pandas>=1.0, <1.4",  # with 1.4.0 the interpolation in bbq_tools fails, pinning for now
    "scipy>=1.5.0",
    "scikit-learn>=0.23.0",
    "tfs-pandas>=3.1",
    "generic-parser>=1.1.0",
    "sdds>=0.1.3",
    "h5py>=2.9.0",
    "tables>=3.6.0",
    "uncertainties>=3.1.4",
    "optics-functions>=0.1.0",
    "turn_by_turn>=0.4.0",
]

# Extra dependencies
EXTRA_DEPENDENCIES = {
    "cern": [
        "jpype1>=1.3.0",
        "pytimber>=2.8.0",
    ],
    "test": [
        "pytest>=5.2",
        "pytest-cov>=2.7",
        "pytest-timeout>=1.4",
        "hypothesis>=5.0.0",
        "jpype1>=1.3.0",
        "attrs>=19.2.0",
    ],
    "doc": [
        "sphinx",
        "travis-sphinx",
        "sphinx_rtd_theme",
    ],
}
EXTRA_DEPENDENCIES.update(
    {"all": [elem for list_ in EXTRA_DEPENDENCIES.values() for elem in list_]}
)


setuptools.setup(
    name=ABOUT_OMC3["__title__"],
    version=ABOUT_OMC3["__version__"],
    description=ABOUT_OMC3["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=ABOUT_OMC3["__author__"],
    author_email=ABOUT_OMC3["__author_email__"],
    url=ABOUT_OMC3["__url__"],
    packages=setuptools.find_packages(exclude=["tests*", "doc"]),
    include_package_data=True,
    python_requires=">=3.7",
    license=ABOUT_OMC3["__license__"],
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    install_requires=DEPENDENCIES,
    tests_require=EXTRA_DEPENDENCIES["test"],
    extras_require=EXTRA_DEPENDENCIES,
)



"""
SuperKEKB
---------

Accelerator-Class for the ``SuperKEKB`` machine.


Model Creation Keyword Args:
    *--Required--*

    - **ring** *(str)*:

        HER or LER ring.

        choices: ``('ler', 'her')``


    *--Optional--*

    - **dpp** *(float)*:

        Deltap/p to use.

        default: ``0.0``


    - **driven_excitation** *(str)*:

        Denotes driven excitation by `AC-dipole` (acd) or by `ADT` (adt)

        choices: ``('acd', 'adt')``


    - **drv_tunes** *(float)*:

        Driven tunes without integer part.


    - **energy** *(float)*:

        Energy in **Tev**.


    - **model_dir** *(str)*:

        Path to model directory; loads tunes and excitation from model!


    - **modifiers** *(str)*:

        Path to the optics file to use (modifiers file).


    - **nat_tunes** *(float)*:

        Natural tunes without integer part.


    - **xing**:

        If True, x-ing angles will be applied to model

        action: ``store_true``
"""
from generic_parser import EntryPoint

from omc3.model.accelerators.accelerator import (Accelerator,
                                                 AcceleratorDefinitionError)
from omc3.utils import logging_tools

LOGGER = logging_tools.get_logger(__name__)


class SKekB(Accelerator):
    """KEK's SuperKEKB accelerator."""
    NAME = "skekb"
    RINGS = ("ler", "her")

    @classmethod
    def get_parameters(cls):
        params = super(SKekB, SKekB).get_parameters()
        params.add_parameter(name="ring", type=str, choices=cls.RINGS, required=True,
                             help="HER or LER ring.")
        return params

    def __init__(self, *args, **kwargs):
        parser = EntryPoint(self.get_parameters(), strict=True)
        opt = parser.parse(*args, **kwargs)
        super().__init__(opt)
        self.ring = opt.ring
        ring_to_beam_direction = {"ler": 1, "her": -1}
        self.beam_direction = ring_to_beam_direction[self.ring]

    @property
    def ring(self):
        if self._ring is None:
            raise AcceleratorDefinitionError("The accelerator definition is incomplete, ring "
                                             "has to be specified (--ring option missing?).")
        return self._ring

    @ring.setter
    def ring(self, value):
        if value not in self.RINGS:
            raise AcceleratorDefinitionError("Ring parameter has to be one of ('ler', 'her')")
        self._ring = value

    def verify_object(self):
        if self.model_dir is None:  # is the class is used to create full response?
            raise AcceleratorDefinitionError("SuperKEKB doesn't have a model creation, "
                                             "calling it this way is most probably wrong.")



"""
Isolation Forest
----------------

This module contains the isolation forest functionality of ``optics_measurements``.
It provides functions to detect and exclude BPMs with anomalies.
"""
import numpy as np
import pandas as pd
from sklearn.ensemble import IsolationForest

from omc3.definitions.constants import PLANE_TO_NUM
from omc3.utils import logging_tools

LOGGER = logging_tools.get_logger(__name__)
ARCS_CONT = 0.01
IRS_CONT = 0.025


def clean_with_isolation_forest(input_files, meas_input, plane):
    bad_bpms = identify_bad_bpms(meas_input, input_files, plane)
    input_files = remove_bad_bpms(input_files, list(set(bad_bpms.NAME)), plane)
    LOGGER.info(str(list(set(bad_bpms.NAME))))
    # TODO potentially write output files ... currently not unique indices!
    #  tfs.write(os.path.join(meas_input.outputdir, f"bad_bpms_iforest_{plane.lower()}.tfs"), bad_bpms)
    return input_files


def identify_bad_bpms(meas_input, input_files, plane):
    bpm_data = pd.concat([tfs_df[["NAME", f"TUNE{plane}", "NOISE_SCALED", f"AMP{plane}"]]
                          for tfs_df in input_files])
    arc_bpm_data, ir_bpm_data = get_data_for_clustering(bpm_data, plane, meas_input.accelerator)
    return pd.concat([identify_single_cluster_bad_bpms(bpm_data, ARCS_CONT, arc_bpm_data, plane),
                      identify_single_cluster_bad_bpms(bpm_data, IRS_CONT, ir_bpm_data, plane)])


def identify_single_cluster_bad_bpms(bpm_tfs_data, cont, data_for_clustering, plane):
    bad_bpms, good_bpms, bad_bpms_scores = detect_anomalies(cont, data_for_clustering, plane)
    bpm_tfs_data, data_for_clustering, bad_bpms, good_bpms = \
        [reassign_index(data) for data in (bpm_tfs_data, data_for_clustering, bad_bpms, good_bpms)]
    signif_feature = get_significant_features(bpm_tfs_data, data_for_clustering, bad_bpms,
                                              good_bpms, plane)
    signif_feature.loc[:, "SCORE"] = bad_bpms_scores
    return signif_feature


def reassign_index(data):
    data["NEW_INDEX"] = range(len(data.NAME))
    return data.set_index("NEW_INDEX")


def get_significant_features(bpm_tfs_data, data_for_clustering, bad_bpms, good_bpms, plane):
    features_df = pd.DataFrame(index=bad_bpms.index)
    for index in bad_bpms.index:
        max_dist = max([(abs(data_for_clustering.loc[index, col] -
                             good_bpms.loc[:, col].mean()), col)
                        for col in [f"TUNE{plane}", "NOISE_SCALED", f"AMP{plane}"]])
        max_dist, sig_col = max_dist
        features_df.loc[index, "NAME"] = bad_bpms.loc[index, "NAME"]
        features_df.loc[index, "FEATURE"] = sig_col
        features_df.loc[index, "VALUE"] = bpm_tfs_data.loc[index, sig_col]
        features_df.loc[index, "AVG"] = np.mean(bpm_tfs_data.loc[good_bpms.index][sig_col])
    return features_df


def detect_anomalies(contamination, data, plane):
    iforest = IsolationForest(n_estimators=100, max_samples='auto',
                              contamination=contamination, max_features=1.0,
                              bootstrap=False)
    features = data[[f"TUNE{plane}", "NOISE_SCALED", f"AMP{plane}"]]
    iforest.fit(features)
    labels = iforest.predict(features)
    bad_bpms = data.iloc[np.where(labels == -1)].copy()
    good_bpms = data.iloc[np.where(labels != -1)].copy()
    bad_bpms_scores = iforest.decision_function(features.iloc[np.where(labels == -1)])
    return bad_bpms, good_bpms, bad_bpms_scores


def get_data_for_clustering(bpm_tfs_data, plane, accelerator):
    arc_bpm_mask = accelerator.get_element_types_mask(bpm_tfs_data.NAME, types=["arc_bpm"])
    ir_bpm_data_for_clustering = bpm_tfs_data.iloc[~arc_bpm_mask].copy()
    arc_bpm_data_for_clustering = bpm_tfs_data.iloc[arc_bpm_mask].copy()
    for col in [f"TUNE{plane}", "NOISE_SCALED", f"AMP{plane}"]:
        ir_bpm_data_for_clustering.loc[:, col] = _normalize_parameter(ir_bpm_data_for_clustering.loc[:, col])
        arc_bpm_data_for_clustering.loc[:, col] = _normalize_parameter(arc_bpm_data_for_clustering.loc[:, col])
    return arc_bpm_data_for_clustering, ir_bpm_data_for_clustering


def _normalize_parameter(column_data):
    return (column_data - column_data.min()) / (column_data.max() - column_data.min())


def remove_bad_bpms(tfs_dfs, bad_bpm_names, plane):
    for i in range(len(tfs_dfs)):
        tfs_dfs[i] = tfs_dfs[i].loc[~tfs_dfs[i].index.isin(bad_bpm_names)]
        tfs_dfs[i].headers[f"Q{PLANE_TO_NUM[plane]}"] = np.mean(tfs_dfs[i][f"TUNE{plane}"])
        tfs_dfs[i].headers[f"Q{PLANE_TO_NUM[plane]}RMS"] = np.std(tfs_dfs[i][f"TUNE{plane}"])
    return tfs_dfs



"""
Phase Advance
-------------

This module contains phase calculation functionality of ``optics_measurements``.
It provides functions to compute betatron phase advances and structures to store them.
"""
from os.path import join
from typing import Dict, Tuple

import numpy as np
import pandas as pd
import tfs
from numpy.typing import ArrayLike

from omc3.optics_measurements.constants import (DELTA, ERR, EXT, MDL, PHASE_NAME, SPECIAL_PHASE_NAME,
                                                TOTAL_PHASE_NAME)
from omc3.optics_measurements.toolbox import ang_sum, df_ang_diff, df_diff
from omc3.utils import logging_tools, stats

LOGGER = logging_tools.get_logger(__name__)


def calculate(
    meas_input: dict, input_files: dict, tunes, plane, no_errors=False
) -> Dict[str, Tuple[Dict[str, tfs.TfsDataFrame], tfs.TfsDataFrame]]:
    """
    Calculate phases for 'free' and 'uncompensated' cases from the measurement files, and return a
    dictionary combining the results for each transverse plane.

    Args:
        meas_input (dict): `OpticsInput` object containing analysis settings from the command-line.
        input_files (dict): `InputFiles` object containing frequency spectra files (linx/y).
        tunes:
        plane:
        no_errors:

    Returns:
        A dictionary of the measured phase advances, with an entry for each horizontal plane. In said entry
        is a dictionary with the measured phase advances for 'free' and 'uncompensated' cases, as well as
        the location of the output ``TfsDataFrames`` for the phases.
    """
    if meas_input.compensation == "none":
        phase_advances, dfs = _calculate_with_compensation(meas_input,
                                                           input_files,
                                                           tunes,
                                                           plane,
                                                           meas_input.accelerator.model,
                                                           'none',
                                                           no_errors)
        uncompensated_phase_advances = phase_advances
    else:
        phase_advances, free_dfs = _calculate_with_compensation(meas_input,
                                                                input_files,
                                                                tunes,
                                                                plane,
                                                                meas_input.accelerator.model,
                                                                meas_input.compensation,
                                                                no_errors)
        LOGGER.info("-- Run uncompensated")
        uncompensated_phase_advances, drv_dfs = _calculate_with_compensation(meas_input,
                                                                             input_files,
                                                                             tunes,
                                                                             plane,
                                                                             meas_input.accelerator.model_driven,
                                                                             'none',
                                                                             no_errors)
        dfs = free_dfs + drv_dfs

    return {'free': phase_advances, 'uncompensated': uncompensated_phase_advances}, dfs



def _calculate_with_compensation(meas_input, input_files, tunes, plane, model_df, compensation='none', no_errors=False):
    """
    Calculates phase advances.

    Args:
        meas_input: the input object including settings and the accelerator class.
        input_files: includes measurement tfs.
        tunes: `TunesDict` object containing measured and model tunes and ac2bpm object
        plane: marking the horizontal or vertical plane, **X** or **Y**.
        no_errors: if ``True``, measured errors shall not be propagated (only their spread).

    Returns:
        A `dictionary` of `TfsDataFrames` indexed (BPMi x BPMj) yielding phase advance `phi_ij`.

         - "MEAS": measured phase advances,
         - "ERRMEAS": errors of measured phase advances,
         - "MODEL": model phase advances.

        +------++--------+--------+--------+--------+
        |      ||  BPM1  |  BPM2  |  BPM3  |  BPM4  |
        +======++========+========+========+========+
        | BPM1 ||   0    | phi_12 | phi_13 | phi_14 |
        +------++--------+--------+--------+--------+
        | BPM2 || phi_21 |    0   | phi_23 | phi_24 |
        +------++--------+--------+--------+--------+
        | BPM3 || phi_31 | phi_32 |   0    | phi_34 |
        +------++--------+--------+--------+--------+
        | BPM4 || phi_41 | phi_42 | phi_43 |    0   |
        +------++--------+--------+--------+--------+

        The phase advance between BPM_i and BPM_j can be obtained via:
        phase_advances["MEAS"].loc[BPMi,BPMj]
        list of output data frames(for files)
    """
    LOGGER.info("Calculating phase advances")
    LOGGER.info(f"Measured tune in plane {plane} = {tunes[plane]['Q']}")

    df = model_df.loc[:, ["S", f"MU{plane}"]]
    how = 'outer' if meas_input.union else 'inner'
    dpp_value = meas_input.dpp if "dpp" in meas_input.keys() else 0
    df = pd.merge(df, input_files.joined_frame(plane, [f"MU{plane}", f"{ERR}MU{plane}"],
                                               dpp_value=dpp_value, how=how),
                  how='inner', left_index=True, right_index=True)
    df[input_files.get_columns(df, f"MU{plane}")] = (input_files.get_data(df, f"MU{plane}")
                                                     * meas_input.accelerator.beam_direction
                                                     )
    phases_mdl = df.loc[:, f"MU{plane}"].to_numpy()
    phase_advances = {"MODEL": _get_square_data_frame(_get_all_phase_diff(phases_mdl), df.index)}
    if compensation == "model":
        df = _compensate_by_model(input_files, meas_input, df, plane)
    phases_meas = input_files.get_data(df, f"MU{plane}")
    if meas_input.compensation == "equation":
        phases_meas = _compensate_by_equation(phases_meas, plane, tunes)

    phases_errors = input_files.get_data(df, f"{ERR}MU{plane}")
    if phases_meas.ndim < 2:
        phase_advances["MEAS"] = _get_square_data_frame(_get_all_phase_diff(phases_meas), df.index)
        phase_advances["ERRMEAS"] = _get_square_data_frame(
                np.zeros((len(phases_meas), len(phases_meas))), df.index)
        return phase_advances
    if meas_input.union:
        mask = np.isnan(phases_meas)
        phases_meas[mask], phases_errors[mask] = 0.0, np.inf
        if no_errors:
            phases_errors[~mask] = 1e-10
    elif no_errors:
        phases_errors = None
    phases_3d = phases_meas[np.newaxis, :, :] - phases_meas[:, np.newaxis, :]
    if phases_errors is not None:
        errors_3d = phases_errors[np.newaxis, :, :] + phases_errors[:, np.newaxis, :]
    else:
        errors_3d = None
    phase_advances["MEAS"] = _get_square_data_frame(stats.circular_mean(
            phases_3d, period=1, errors=errors_3d, axis=2) % 1.0, df.index)
    phase_advances["ERRMEAS"] = _get_square_data_frame(stats.circular_error(
            phases_3d, period=1, errors=errors_3d, axis=2), df.index)
    return phase_advances, [_create_output_df(phase_advances, df, plane),
                            _create_output_df(phase_advances, df, plane, tot=True)]


def _compensate_by_equation(phases_meas: ArrayLike, plane, tunes):
    driven_tune, free_tune, ac2bpmac = tunes[plane]["Q"], tunes[plane]["QF"], tunes[plane]["ac2bpm"]
    k_bpmac = ac2bpmac[2]
    phase_corr = ac2bpmac[1] - phases_meas[k_bpmac] + (0.5 * driven_tune)
    phases_meas = phases_meas + phase_corr[np.newaxis, :]
    r = tunes.get_lambda(plane)
    phases_meas[k_bpmac:, :] = phases_meas[k_bpmac:, :] - driven_tune
    psi = (np.arctan((1 - r) / (1 + r) * np.tan(2 * np.pi * phases_meas)) / (2 * np.pi)) % 0.5
    phases_meas = np.where(phases_meas % 1.0 > 0.5, psi + .5, psi)
    phases_meas[k_bpmac:, :] = phases_meas[k_bpmac:, :] + free_tune
    return phases_meas


def _compensate_by_model(input_files, meas_input, df, plane):
    df = pd.merge(df, pd.DataFrame(meas_input.accelerator.model_driven.loc[:, [f"MU{plane}"]]),
                  how='inner', left_index=True, right_index=True, suffixes=("", "comp"))
    phase_compensation = df_diff(df, f"MU{plane}", f"MU{plane}comp")
    df[input_files.get_columns(df, f"MU{plane}")] = ang_sum(
        input_files.get_data(df, f"MU{plane}"), phase_compensation[:, np.newaxis])
    return df


def write(dfs, headers, output, plane):
    LOGGER.info(f"Writing phases: {len(dfs)}")
    for head, df, name in zip(headers, dfs, (PHASE_NAME, TOTAL_PHASE_NAME, PHASE_NAME+"driven_", TOTAL_PHASE_NAME+"driven_")):
        tfs.write(join(output, f"{name}{plane.lower()}{EXT}"), df, head)
        LOGGER.info(f"Phase advance beating in {name}{plane.lower()}{EXT} = "
                    f"{stats.weighted_rms(df.loc[:, f'{DELTA}PHASE{plane}'])}")


def _create_output_df(phase_advances, model, plane, tot=False):
    meas = phase_advances["MEAS"]
    mod = phase_advances["MODEL"]
    err = phase_advances["ERRMEAS"]
    if tot:
        output_data = model.loc[:, ["S", f"MU{plane}"]].iloc[:, :]
        output_data["NAME"] = output_data.index
        output_data = output_data.assign(S2=model.at[model.index[0], "S"], NAME2=model.index[0])
        output_data[f"PHASE{plane}"] = meas.to_numpy()[0, :]
        output_data[f"{ERR}PHASE{plane}"] = err.to_numpy()[0, :]
        output_data[f"PHASE{plane}{MDL}"] = mod.to_numpy()[0, :]
    else:
        output_data = model.loc[:, ["S", f"MU{plane}"]].iloc[:-1, :]
        output_data["NAME"] = output_data.index
        output_data = output_data.assign(S2=model.loc[:, "S"].to_numpy()[1:], NAME2=model.index[1:].to_numpy())
        output_data[f"PHASE{plane}"] = np.diag(meas.to_numpy(), k=1)
        output_data[f"{ERR}PHASE{plane}"] = np.diag(err.to_numpy(), k=1)
        output_data[f"PHASE{plane}{MDL}"] = np.diag(mod.to_numpy(), k=1)
    output_data.rename(columns={f"MU{plane}": f"MU{plane}{MDL}"}, inplace=True)
    output_data[f"{DELTA}PHASE{plane}"] = df_ang_diff(output_data, f"PHASE{plane}", f"PHASE{plane}{MDL}")
    output_data[f"{ERR}{DELTA}PHASE{plane}"] = output_data.loc[:, f"{ERR}PHASE{plane}"].to_numpy()
    return output_data


def _get_all_phase_diff(phases_a: ArrayLike, phases_b: ArrayLike = None):
    if phases_b is None:
        phases_b = phases_a
    return (phases_a[np.newaxis, :] - phases_b[:, np.newaxis]) % 1.0


def _get_square_data_frame(data, index):
    return pd.DataFrame(data=data, index=index, columns=index)


def write_special(meas_input, phase_advances, plane_tune, plane):
    # TODO REFACTOR AND SIMPLIFY
    accel = meas_input.accelerator
    meas = phase_advances["MEAS"]
    bd = accel.beam_direction
    elements = accel.elements
    special_phase_columns = ['ELEMENT1',
                             'ELEMENT2',
                             f'PHASE{plane}',
                             f'{ERR}PHASE{plane}',
                             f'PHASE{plane}_DEG',
                             f'{ERR}PHASE{plane}_DEG',
                             f'PHASE{plane}{MDL}',
                             f'PHASE{plane}{MDL}_DEG',
                             'BPM1',
                             'BPM2',
                             f'BPM_PHASE{plane}',
                             f'BPM_{ERR}PHASE{plane}',]
    special_phase_df = pd.DataFrame(columns=special_phase_columns)
    
    for elem1, elem2 in accel.important_phase_advances():
        mus1 = elements.loc[elem1, f"MU{plane}"] - elements.loc[:, f"MU{plane}"]
        minmu1 = abs(mus1.loc[meas.index]).idxmin()
        mus2 = elements.loc[:, f"MU{plane}"] - elements.loc[elem2, f"MU{plane}"]
        minmu2 = abs(mus2.loc[meas.index]).idxmin()
        bpm_phase_advance = meas.loc[minmu1, minmu2]
        model_value = elements.loc[elem2, f"MU{plane}"] - elements.loc[elem1, f"MU{plane}"]
        if (elements.loc[elem1, "S"] - elements.loc[elem2, "S"]) * bd > 0.0:
            bpm_phase_advance += plane_tune
            model_value += plane_tune
        bpm_err = phase_advances["ERRMEAS"].loc[minmu1, minmu2]
        elems_to_bpms = -mus1.loc[minmu1] - mus2.loc[minmu2]
        ph_result = ((bpm_phase_advance + elems_to_bpms) * bd)
        model_value = (model_value * bd) % 1
        special_phase_df=special_phase_df.append(dict(zip(special_phase_columns,[
                                                            elem1,
                                                            elem2,
                                                            ph_result % 1,
                                                            bpm_err,
                                                            _to_deg(ph_result),
                                                            bpm_err * 360,
                                                            model_value,
                                                            _to_deg(model_value),
                                                            minmu1,
                                                            minmu2,
                                                            bpm_phase_advance,
                                                            elems_to_bpms,
        ])), ignore_index=True)

    tfs.write(join(meas_input.outputdir, f"{SPECIAL_PHASE_NAME}{plane.lower()}{EXT}"), special_phase_df)


def _to_deg(phase):  # -90 to 90 degrees
    phase = phase % 0.5 * 360
    if phase < 90:
        return phase
    return phase - 180



"""
Plotting Utilities: Style
-------------------------

Helper functions to style plots.
"""
from typing import Dict, Any, Sequence, Union

import matplotlib
from generic_parser.entry_datatypes import get_instance_faker_meta
from matplotlib import pyplot as plt
from pathlib import Path

from omc3.utils.iotools import PathOrStr

REMOVE_ENTRY = "REMOVE ENTRY"  # id to remove entries in manual style

STYLES_DIR = Path(__file__).parent.parent / 'styles'


# Style ######################################################################


def omc3_styles():
    return {p.with_suffix('').name: p for p in STYLES_DIR.glob('*.mplstyle')}


def set_style(styles: Union[Path, str, Sequence[Union[Path, str]]] = 'standard',
              manual: Dict[str, Any] = None):
    """
    Sets the style for all following plots.

    Args:
        styles: `List` of styles (or single string), either path to style-file, name of style in
            styles or from the mpl styles
        manual: `Dict` of manual parameters to update. Convention: ``REMOVE_ENTRY`` removes entry
                from given styles, i.e. falls back to mpl default.
    """
    if isinstance(styles, PathOrStr):
        styles = (styles,)

    local_styles = omc3_styles()
    styles = [local_styles.get(style, style) for style in styles]

    if manual:
        for key, value in manual.items():
            if value == REMOVE_ENTRY:
                manual[key] = matplotlib.rcParams[key]
        styles.append(manual)
    plt.style.use(styles)
